from functools import lru_cache
from typing import Dict, Tuple, Union, List
import os
from transformers import AutoTokenizer, AutoConfig, AutoModel
from transformers.optimization import get_linear_schedule_with_warmup, get_cosine_schedule_with_warmup
from pytorch_lightning import LightningModule
import torch
import tempfile
import json
from omegaconf import OmegaConf, DictConfig


def get_hf_tokenizer(config: Union[Dict, DictConfig] , vocab: Union[Dict, DictConfig]):
    with tempfile.TemporaryDirectory() as tmpdirname:
        vocab_path = os.path.join(tmpdirname, 'vocab.txt')
        config_path = os.path.join(tmpdirname, 'config.json')
        with open(vocab_path, 'w') as f:
            for k in vocab.keys():
                f.writelines(k + '\n')
        with open(config_path, 'w') as f:
            if type(config) == DictConfig:
                config = OmegaConf.to_container(config)
            assert type(config)==dict, f'config must be type dict, but found {type(config)}'
            d = json.dumps(config)
            f.write(d)
        tokenizer = AutoTokenizer.from_pretrained(tmpdirname)
    return tokenizer

def get_hf_config_object(config: Union[DictConfig , Dict]):
    with tempfile.TemporaryDirectory() as tmpdirname:
        config_path = os.path.join(tmpdirname, 'config.json')
        with open(config_path, 'w') as f:
            if type(config) == DictConfig:
                config = OmegaConf.to_container(config)
            try:
                config = config.to_dict() # 兼容huffingface config 对象
            except:
                pass
            assert type(config)==dict, f'config must be type dict, but found {type(config)}'
            d = json.dumps(config)
            f.write(d)
        config = AutoConfig.from_pretrained(tmpdirname)
    return config

    

def align_token_span(token_span_offset: Tuple, token_offset_mapping: List[Tuple]) -> Tuple:
    '''将词符级别的下标对齐为字符级别的下标
    参数
    - token_span_offset: 例如(0, 1) 下标指的是字符的下标
    - token_offset_mapping: 每个词符与字符对应的下标[(0,1),(1,2)]
    返回
    char_span_offset: (0,2)
    '''
    char_span_offset = ()
    if token_span_offset[1] - token_span_offset[0] == 1:
        char_span_offset = token_offset_mapping[token_span_offset[0]]
        return char_span_offset

    else:
        start = token_offset_mapping[token_span_offset[0]][0]
        end = token_offset_mapping[token_span_offset[1]-1][1]
        char_span_offset = (start, end)
        return char_span_offset
    
    
class PLMBaseModel(LightningModule):
    """基于预训练语音模型的基类,在继承类的时候需要传入plm和plm_dir
    
    - 内置了scheduler,可以通过cls.scheduler_names查看所有的scheduler,通过self.get_scheduler_config方法得到pl的scheduler config
    - 通过self.tokenizer直接调用tokenizer
    - 通过self.get_plm_architecture可以得到预训练模型的架构,主要并没有加载预训练参数
    """
    
    scheduler_names = ['linear_warmup_step', 'cosine_warmup_step', 'harmonic_epoch', '1cycle_step']
    
    def __init__(self) -> None:
        super().__init__()
        self.save_hyperparameters()
        assert 'plm' in self.hparams and 'plm_dir' in self.hparams, 'you have to at least pass in plm and plm_dir'
        if 'label2id' not in self.hparams and 'id2label' in self.hparams:
            self.hparams.label2id = {l:i for i,l in self.hparams.id2label.items()}
        
    
    @property
    @lru_cache()
    def tokenizer(self):
        keys = self.hparams.keys()
        if 'trf_config' in keys and 'vocab' in keys:
            return get_hf_tokenizer(self.hparams.trf_config, self.hparams.vocab)
        elif 'plm' in keys and 'plm_dir' in keys:
            plm_path = os.path.join(self.hparams.plm_dir, self.hparams.plm)
            return AutoTokenizer.from_pretrained(plm_path)


    def get_plm_config(self):
        if 'trf_config' in self.hparams.keys():
            return get_hf_config_object(self.hparams.trf_config)
        elif 'plm' in self.hparams.keys() and 'plm_dir' in self.hparams.keys():
            plm_path = os.path.join(self.hparams.plm_dir, self.hparams.plm)
            plm_config = AutoConfig.from_pretrained(plm_path)    
            return plm_config
    
    
    def get_plm_architecture(self, add_pooler_layer: bool = False):
        if 'trf_config' in self.hparams.keys():
            plm_config = get_hf_config_object(self.hparams.trf_config)
        elif 'plm' in self.hparams.keys() and 'plm_dir' in self.hparams.keys():
            plm_path = os.path.join(self.hparams.plm_dir, self.hparams.plm)
            plm_config = AutoConfig.from_pretrained(plm_path)
            self.hparams.trf_config = plm_config.to_dict()
            self.hparams.vocab = dict(sorted(self.tokenizer.vocab.items(), key=lambda x: x[1]))
        plm_config.add_pooler_layer = add_pooler_layer
        return AutoModel.from_config(plm_config)    
    
    
    def get_linear_warmup_step_scheduler_config(self, optimizer, warmup_rate: float = 0.5):
        steps_per_epoch = self.get_one_epoch_steps()
        total_steps = self.trainer.max_epochs * steps_per_epoch
        scheduler = get_linear_schedule_with_warmup(optimizer=optimizer, num_training_steps=total_steps, num_warmup_steps=warmup_rate * steps_per_epoch)
        scheduler_config = {'scheduler': scheduler, 'interval':'step'}
        return scheduler_config
    
    
    def get_cosine_warmup_step_scheduler_config(self, optimizer, warmup_rate: float = 0.5):
        steps_per_epoch = self.get_one_epoch_steps()
        total_steps = self.trainer.max_epochs * steps_per_epoch
        scheduler = get_cosine_schedule_with_warmup(optimizer=optimizer, num_training_steps=total_steps, num_warmup_steps=warmup_rate * steps_per_epoch)
        scheduler_config = {'scheduler': scheduler, 'interval':'step'}
        return scheduler_config
    
    
    def get_harmonic_epoch_scheduler_config(self, optimizer):
        scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, lambda epoch: 1.0 / (epoch + 1.0))
        scheduler_config = {'scheduler': scheduler, 'interval':'epoch'}
        return scheduler_config
    
    
    def get_one_cycle_scheduler_config(self, optimizer):
        steps_per_epoch = self.get_one_epoch_steps()
        total_steps = self.trainer.max_epochs * steps_per_epoch
        scheduler = torch.optim.lr_scheduler.OneCycleLR(optimizer=optimizer, max_lr=self.hparams.lr*3, total_steps=total_steps)
        scheduler_config = {'scheduler': scheduler, 'interval':'step'}
        return scheduler_config
    
    
    @property
    def epoch_schedulers(self):
        return [scheduler for scheduler in self.scheduler_names if scheduler.endswith('epoch')]


    @property
    def step_schedulers(self):
        return [scheduler for scheduler in self.scheduler_names if scheduler.endswith('step')]
    
    
    def get_one_epoch_steps(self):
        return self.trainer.estimated_stepping_batches
        
    
    def get_scheduler_config(self, optimizer, name: str):
        """

        Args:
            optimizer (): 优化器实例
            name (str): 'harmonic_epoch', 'linear_warmup_step', 'cosine_warmup_step'之一

        Returns:
            dict: scheduler配置字典
        """
        availabel_names = self.scheduler_names
        assert name in availabel_names, f'availabel names {availabel_names}'
        if name == 'harmonic_epoch':
            return self.get_harmonic_epoch_scheduler_config(optimizer=optimizer)
        elif name == 'linear_warmup_step':
            return self.get_linear_warmup_step_scheduler_config(optimizer=optimizer)
        elif name == 'cosine_warmup_step':
            return self.get_cosine_warmup_step_scheduler_config(optimizer=optimizer)
        elif name == '1cycle':
            return self.get_one_cycle_scheduler_config(optimizer=optimizer)

    
    def to_onnx(self, 
                file_path: str, 
                text_a: str = '中国人', 
                text_b : str = '中国'):
        torch_inputs = self.tokenizer(text_a, text_b, return_tensors='pt')
        dynamic_axes = {
                    'input_ids': {0: 'batch', 1: 'seq'},
                    'attention_mask': {0: 'batch', 1: 'seq'},
                    'token_type_ids': {0: 'batch', 1: 'seq'},
                }
        with torch.no_grad():
            torch.onnx.export(model=self,
                              args=tuple(torch_inputs.values()), 
                              f=file_path, 
                              input_names=list(torch_inputs.keys()),
                              dynamic_axes=dynamic_axes, 
                              opset_version=14,
                              output_names=['logits'],
                              export_params=True)
        print('export to onnx successfully')


from .span_classifier import get_spancat
from .token_classifier import get_tokencat
from .text_classifier import get_text_classification
from .sentence_segmenter import get_sentence_segmenter
from .relation_extractor import get_relation_extractor
from .information_extractor import get_information_extractor
from .entity_normalizer import get_entity_normalizer


from datasets import Dataset, DatasetDict, load_from_disk
from typing import Dict, Tuple, Union, List
from ..algorithms.text_match import BM25
from spacy.tokens import Doc
import random
from tqdm import tqdm


def train_val_split(dataset: Dataset, 
                    val_frac: float =0.1,
                    return_dataset_dict: bool =True) -> Union[Tuple[Dataset, Dataset], DatasetDict]:
    """split dataset into tarin and validation datasets

    Args:
        dataset (Dataset): dataset to split
        val_frac (float, optional): validation radio of all dataset. Defaults to 0.1.
        return_dataset_dict (bool, optional): if return_dataset_dict is True, return a DatasetDict,
            otherwise return a tuple of train, val datasets. Defaults to True.

    Returns:
        Union[Tuple[Dataset, Dataset], DatasetDict]: if return_dataset_dict is True, return a DatasetDict, otherwise return a tuple of train, val datasets
    """
    df = dataset.to_pandas()
    train_df = df.sample(frac=1-val_frac)
    val_df = df.drop(train_df.index)
    train_ds = Dataset.from_pandas(train_df, preserve_index=False)
    val_ds = Dataset.from_pandas(val_df, preserve_index=False)
    if not return_dataset_dict:
        return train_ds, val_ds
    else:
        return DatasetDict({'train': train_ds, 'validation': val_ds})


def train_val_test_split(dataset: Dataset,
                         val_frac: float =0.1,
                         test_frac: float =0.1,
                         return_dataset_dict: bool =True) -> Union[Tuple[Dataset, Dataset, Dataset], DatasetDict]:
    """split dataset into tarin vlidation and test datasets

    Args:
        dataset (Dataset): dataset to split
        val_frac (float, optional): validation radio of all dataset. Defaults to 0.1.
        test_frac (float, optional): test radio of all dataset. Defaults to 0.1.

    Returns:
        Union[Tuple[Dataset, Dataset, Dataset], DatasetDict]: if return_dataset_dict is True, return a DatasetDict, 
            otherwise return a tuple of train, val, test datasets
        
    """
    df = dataset.to_pandas()
    train_df = df.sample(frac=1-val_frac-test_frac)
    other_df = df.drop(train_df.index)
    val_df = other_df.sample(frac=1-(test_frac/(val_frac+test_frac)))
    test_df = other_df.drop(val_df.index)
    train_ds = Dataset.from_pandas(train_df, preserve_index=False)
    val_ds = Dataset.from_pandas(val_df, preserve_index=False)
    test_ds = Dataset.from_pandas(test_df, preserve_index=False)
    if not return_dataset_dict:
        return train_ds, val_ds, test_ds
    else:
        return DatasetDict({'train': train_ds, 'validation': val_ds, 'test': test_ds})
        

def make_text_match_dataset_with_bm25(corpus: List[str],
                                     synonym_dict: Dict[str, set],
                                     num_positive_samples: int=10,
                                     num_negative_samples: int=20,
                                     recall_topk: int = 1000,
                                     reverse_sample: bool = False,
                                     positive_label: str = '1',
                                     negative_label: str='0',
                                     tokenizer = None,
                                     k1: float = 1.5,
                                     b: float=0.75,
                                     epsilon: float=0.25,
                                     is_retrain_docs=True,
                                     return_bm25: bool = False):
    """制作文本匹配二分类数据集, 此数据集适用于text_pair_classification任务

    Args:
        corpus (List[str]): 所有的可以召回的语料,词语等
        synonym_dict (Dict[str, set]): 标准词对应的同义词字典.字典的键应该包含于corpus.
        num_negative_samples (int): 负样本的数量.
        recall_topk(int): 召回模型的召回数量
        reverse_sample (bool): 是否逆序负样本采样
        positive_label (str): 正样本标签, 默认为'1'.
        negative_label (str): 负样本标签, 默认为'0'.
        tokenizer (_type_, optional): 分词器,默认为字符切分. Defaults to None.
        k1 (float, optional): bm25参数. Defaults to 1.5.
        b (float, optional): bm25参数. Defaults to 0.75.
        epsilon (float, optional): bm25参数. Defaults to 0.25.
        is_retrain_docs (bool, optional): bm25参数, 是否保留文档. Defaults to True.
        return_bm25 (bool): 是否返回bm25模型. Defaults to False.
    """
    bm25 = BM25(corpus=corpus, 
                k1=k1, 
                b=b, 
                epsilon=epsilon,
                is_retain_docs=is_retrain_docs,
                tokenizer=tokenizer)
    label_ls = []
    text_a_ls = []
    text_b_ls = []
    for key in tqdm(synonym_dict.keys()):
        recalls = bm25.recall(key, topk=recall_topk)
        recalls = [r[0] for r in recalls if r[0] != key and r[0] not in synonym_dict[key]]
        if reverse_sample:
            recalls.reverse()
            negatives = recalls[:num_negative_samples]
        else:
            negatives = recalls[:num_negative_samples]
        text_a_ls.extend(negatives)
        text_b_ls.extend([key]*len(negatives))
        label_ls.extend([negative_label]*len(negatives))
        syms = list(synonym_dict[key])
        positives = random.choices(syms, k=num_positive_samples)
        text_a_ls.extend(positives)
        text_b_ls.extend([key]*num_positive_samples)
        label_ls.extend([positive_label]*num_positive_samples)
    ds = Dataset.from_dict({'text_a':text_a_ls, 'text_b':text_b_ls, 'label': label_ls})
    if not return_bm25:
        return ds
    else:
        return ds, bm25
    

def make_relation_extraction_dataset_from_docs(docs: List[Doc], 
                                              prompt_type: bool=False,
                                              num_negative: int = 2): 
    """将doc._.relations 转换为relation extraction任务数据

    Args:
        docs (List[Doc]): 待转换docs
        prompt_type (bool, optional): 是否为prompt形式. Defaults to False.

    Returns:
        Dataset: all convert dataset
        
    Example:
        dataset: 
        {'text': '小张毕业于河北工程大学',
        'triples': [{'object': {'offset': [5,11], 'text': '河北工程大学'},
                    'predicate': '毕业院校',
                    'subject': {'offset': [0, 2], 'text': '小张'}}]},
    """
    if not prompt_type:
        text_ls = []
        triple_ls = []
        for doc in tqdm(docs):
            text_ls.append(doc.text)
            triples = []
            for rel in doc._.relations:
                sub = rel.sub
                pred = rel.label
                for obj in rel.objs:
                    triples.append({'subject':{'offset':(sub.start_char, sub.end_char), 'text':sub.text},
                                    'predicate': pred,
                                    'object':{'offset':(obj.start_char, obj.end_char), 'text':obj.text}})
            triple_ls.append(triples)
        ds = Dataset.from_dict({'text':text_ls, 'triples':triple_ls})
        return ds
    else:
        text_ls = []
        triple_ls = []
        prompt_ls = []
        all_labels = set([rel.label for doc in docs for rel in doc._.relations])
        for doc in tqdm(docs):
            triple_dict = {}
            for rel in doc._.relations:
                sub = rel.sub
                p = rel.label
                if p not in triple_dict:
                    triple_dict[p] = []
                for obj in rel.objs:
                    triple_dict[p].append({'subject':{'offset':(sub.start_char, sub.end_char), 'text':sub.text},
                                    'object':{'offset':(obj.start_char, obj.end_char), 'text':obj.text}})
            
            for p in triple_dict:
                text_ls.append(doc.text)
                triple_ls.append(triple_dict[p])
                prompt_ls.append(p)
            other_labels = list(all_labels - triple_dict.keys())
            for l in other_labels[:num_negative]:
                text_ls.append(doc.text)
                triple_ls.append([])
                prompt_ls.append(l)
                
        ds = Dataset.from_dict({'text':text_ls, 'triples':triple_ls, 'prompts': prompt_ls})
        return ds


def make_event_extraction_ds_from_doc(docs: List[Doc]):
    text_ls = []
    event_ls = []
    for doc in tqdm(docs):
        text_ls.append(doc.text)
        events = []
        for event in doc._.events:
            e = {'label':event.label, 'roles':[]}
            for role in event.roles:
                spans = event.roles[role]
                for span in spans:
                    e['roles'].append({'label':role, 
                                    'offset':(span.start_char, span.end_char), 
                                    'text':span.text})
            events.append(e)
        event_ls.append(events)
    ds = Dataset.from_dict({'text':text_ls, 'events':event_ls})
    return ds


def make_entity_extraction_ds_from_doc(docs: List[Doc], 
                                      from_ent_or_span:str = 'ent',
                                      span_key: str = 'all'):
    """制作entity extraction 格式的数据集,只能得到连续的实体

    Args:
        docs (List[Doc]): 含有ent标记的spacy文档
        from_ent_or_span (str): 实体标记,ent为doc.ent,span为doc.spans,当为span时请确保span_key正确
        span_key (str): doc.spans的主键,即doc.spans[span_key]

    Returns:
        ds (Dataset): 数据模块EntityExtraction格式的数据集
    """
    assert from_ent_or_span in ['ent', 'span'], 'only ent or span available'
    text_ls = []
    ent_ls = []
    if from_ent_or_span == 'ent':
        for doc in tqdm(docs):
            text_ls.append(doc.text)
            ents = []
            for ent in doc.ents:
                e = {'label': ent.label_,'indexes':[i for i in range(ent.start_char, ent.end_char)], 'text':ent.text}
                ents.append(e)
            ent_ls.append(ents)
    elif from_ent_or_span == 'span':
        for doc in tqdm(docs):
            text_ls.append(doc.text)
            ents = []
            for span in doc.spans[span_key]:
                e = {'label': span.label_, 'indexes':[i for i in range(span.start_char, span.end_char)], 'text':span.text}
                ents.append(e)
    ds = Dataset.from_dict({'text':text_ls, 'entities':ent_ls})
    return ds  


import torch
from pytorch_lightning import LightningModule
from torchmetrics import F1Score
from typing import List, Any, Dict
from transformers import AutoModel, BertModel, BertConfig, BertTokenizer
import torch.nn.functional as F
import os
from ...layers.classifier import SimpleDense




class BERTCrossEncoder(LightningModule):
    '''
    交互式的bert句子对分类模型
    '''
    def __init__(
        self, 
        lr: float ,
        hidden_size: int,
        dropout: float,
        weight_decay: float ,
        **kwargs
        ):
        super(BERTCrossEncoder, self).__init__()  
        
        self.save_hyperparameters(logger=False)
        # 模型结构
        self.bert = BertModel(self.hparams['trf_config'])
        self.dropout = torch.nn.Dropout(dropout)
        self.classifier = SimpleDense(self.bert.config.hidden_size, hidden_size=hidden_size, output_size=len(self.hparams.label2id))
        
        # 损失函数
        self.criterion = torch.nn.CrossEntropyLoss()
        # 指标评价函数
        self.train_f1 = F1Score(average='macro', num_classes=len(self.hparams.label2id))
        self.val_f1 = F1Score(average='macro', num_classes=len(self.hparams.label2id))
        self.test_f1 = F1Score(average='macro', num_classes=len(self.hparams.label2id))
        # 预处理流程
        self.tokenizer = self._init_tokenizer()
        

    def forward(self, input_ids, token_type_ids, attention_mask):
        x = self.bert(input_ids=input_ids, token_type_ids=token_type_ids, attention_mask=attention_mask)
        x = x.last_hidden_state
        x = x[:, 0] # CLS
        x = self.dropout(x)
        logits = self.classifier(x)
        return logits


    def shared_step(self, batch):
        inputs = batch['inputs']
        labels = batch['label_ids']
        logits = self(**inputs)
        loss = self.criterion(logits, labels)
        preds = torch.argmax(logits, dim=-1)
        return loss, preds, labels

    def training_step(self, batch, batch_idx):
        loss, preds, labels = self.shared_step(batch)
        self.train_f1(preds, labels)
        self.log('train/f1', self.train_f1, on_step=True, on_epoch=True, prog_bar=True)
        return {'loss': loss}

    def validation_step(self, batch, batch_idx):
        loss, preds, labels = self.shared_step(batch)
        self.val_f1(preds, labels)
        self.log('val/f1', self.val_f1, on_step=False, on_epoch=True, prog_bar=True)
        return {'loss': loss}


    def test_step(self, batch, batch_idx):
        loss, preds, labels = self.shared_step(batch)
        self.test_f1(preds, labels)
        self.log('test/f1', self.test_f1, on_step=False, on_epoch=True, prog_bar=True)
        return {'loss': loss}


    def configure_optimizers(self):
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
        grouped_parameters = [
            {'params': [p for n, p in self.bert.named_parameters() if not any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.bert.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': 0.0},
            {'params': [p for n, p in self.classifier.named_parameters() if not any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr *5, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.classifier.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr * 5, 'weight_decay': 0.0}
        ]
        optimizer = torch.optim.AdamW(grouped_parameters)
        scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, lambda epoch: 1.0 / (epoch + 1.0))
        return [optimizer], [scheduler]

    def _init_tokenizer(self):
        with open('./vocab.txt', 'w') as f:
            for k in self.hparams.vocab.keys():
                f.writelines(k + '\n')
        self.hparams.trf_config.to_json_file('./config.json')
        tokenizer = BertTokenizer.from_pretrained('./')
        os.remove('./vocab.txt')
        os.remove('./config.json')
        return tokenizer
    
    def to_onnx(self, file_path: str):
        text1 = '我是中国人'
        text2 = '我是河北人'
        torch_inputs = self.tokenizer(text1, text2, return_tensors='pt')
        dynamic_axes = {
                    'input_ids': {0: 'batch', 1: 'seq'},
                    'attention_mask': {0: 'batch', 1: 'seq'},
                    'token_type_ids': {0: 'batch', 1: 'seq'},
                }
        with torch.no_grad():
            torch.onnx.export(
                model=self,
                args=tuple(torch_inputs.values()), 
                f=file_path, 
                input_names=list(torch_inputs.keys()),
                dynamic_axes=dynamic_axes, 
                opset_version=14,
                output_names=['logits'],
                export_params=True)
        print('export to onnx successfully')
        
        
    def predict(self, text_pair: List, device: str='cpu') -> Dict[str, float]:
        device = torch.device(device)
        inputs = self.tokenizer(
                text_pair[0],
                text_pair[1],
                padding='max_length',
                max_length=self.hparams.max_length,
                return_tensors='pt',
                truncation=True)
        inputs.to(device)
        # self.to(device)
        # self.freeze()
        self.eval()
        with torch.no_grad():
            logits = self(**inputs)
            scores = torch.nn.functional.softmax(logits, dim=-1).tolist()
            cats = {}
            for i, v in enumerate(scores[0]): # scores : [[0.1, 0.2, 0.3, 0.4]]
                id2label = {v: k for k, v in self.hparams.label2id.items()}
                cats[id2label[i]] = v
        return sorted(cats.items(), key=lambda x: x[1], reverse=True)




from torchmetrics import Metric 
from typing import List, Set, Union, Optional
import torch
from collections import namedtuple

Role = namedtuple('Role', ['start', 'end', 'label'])

class Event:
    def __init__(self, 
                 label: Union[str, int, torch.Tensor], 
                 roles: Set[List[Role]]):
        super().__init__()
        self.label = label
        self.roles = roles

    def __hash__(self) -> int:
        return hash(self.label)

    def __eq__(self, __o: object) -> bool:
        return __o.label == self.label and __o.roles == self.roles

    def __repr__(self) -> str:
        return f'Event: <{self.label}>'


class EventF1(Metric):
    """事件抽取评价指标
    """

    full_state_update: Optional[bool] = False

    def __init__(self):
        super().__init__()
        self.add_state('correct', default=torch.tensor(0.0), dist_reduce_fx='sum')
        self.add_state('all_pred', default=torch.tensor(0.0), dist_reduce_fx='sum')
        self.add_state('all_true', default=torch.tensor(0.0), dist_reduce_fx='sum')

    def update(self, pred: List[Set[Event]], true: List[Set[Event]]) -> None:
        """
        参数:
        - pred: batch_size大小的列表, 列表内为每个样本的预测的三元组的集合(set)
        - true: 同pred
        """
        assert len(pred) == len(true), f'pred : {pred}, true : {true}'
        for p, t in zip(pred, true):
            self.correct += len(p & t)
            self.all_pred += len(p)
            self.all_true += len(t)
    
    def compute(self):
        # 下面这个公式可以通过化简得到
        f1 =  2 * self.correct / (self.all_pred + self.all_true + 1e-10)
        return f1


import googletrans
import requests
from googletrans import Translator


def augment_text_from_youdao_translator(text: str):
    """通过调用有道翻译接口增加文本数据,最多可以扩充三条文本

    Args:
        text (str): 原始文本

    Returns:
        List[str]: 增强的文本,不包含原始文本. 
    """
    results = []
    trans_types = [('ZH_CN2EN','EN2ZH_CN'),('ZH_CN2KR','KR2ZH_CN'),('ZH_CN2JA','JA2ZH_CN'),]
    for t in trans_types:
        try:
            data1 = { 'doctype': 'json', 'type': t[0],'i': text }
            r = requests.get("http://fanyi.youdao.com/translate",params=data1)
            r = r.json()
            tgt = r['translateResult'][0][0]['tgt']
            
            data2 = { 'doctype': 'json', 'type': t[1],'i': tgt}
            r = requests.get("http://fanyi.youdao.com/translate",params=data2)
            r = r.json()
            res = r['translateResult'][0][0]['tgt']
            if res not in results and res!=text:
                results.append(res)
        except:
            pass
    return results

def augment_text_from_google_translator(text: str, num_augs: int=2):
    translator = Translator(service_urls=['translate.google.cn'])
    res_ls = []
    langs = list(googletrans.LANGUAGES.keys())
    for lang in langs:
        aug = translator.translate(text, dest=lang, src='zh-cn').text
        res = translator.translate(aug, dest='zh-cn', src=lang).text
        if res not in res_ls and res != text:
            res_ls.append(res)
        if len(res_ls) == num_augs:
            break
    return res_ls
        

if __name__ == "__main__":
    text = '我今天感冒了,有点头疼.'
    res = augment_text_from_youdao_translator(text)
    print("有道:",res)
    res = augment_text_from_google_translator(text, 10)
    print("google:",res)
        


from functools import lru_cache
from ..utils.make_datamodule import char_idx_to_token, get_logger, PLMBaseDataModule
import torch
from typing import Dict, List, Optional, Union
import numpy as np
import pandas as pd
import random


log = get_logger()


class RelationExtractionDataModule(PLMBaseDataModule):
    """三元组抽取数据模块,
        数据集样式:
            {'text':'小明是小花的朋友'。
            'triples': [{'object': {'offset': [0, 2], 'text': '小明'},'predicate': '朋友','subject': {'offset': [3, 5], 'text': '小花'}}]}}
        """
    def __init__(self, 
                 dataset: str, 
                 batch_size: int ,
                 plm: str = 'hfl/chinese-roberta-wwm-ext',
                 transform: str = 'gplinker',
                 auto_length: Union[int, str] = 'max',
                 **kwargs):
        """
        Args:
            dataset (str): 数据集名称
            plm (str): 预训练模型名称
            transform(str): 转换特征格式,可以通过get_available_transforms查看
            auto_length (str, int): 自动设置最大长度的策略, 可以为'max', 'mean'或者>0的数,超过512的设置为512
            batch_size (int): 训练,验证,测试数据集的批次大小,
            num_workers (int): 多进程数
            pin_memory (bool): 是否应用锁页内存,
            dataset_dir (str): 数据集默认路径
            plm_dir (str): 预训练模型默认路径
        """
        super().__init__()
                
        self.transforms = {'gplinker': self.gplinker_transform,
                           'onerel': self.onerel_transform,
                           'casrel': self.casrel_transform}
        assert self.hparams.transform in self.transforms.keys(), f'availabel models for relation extraction: {self.transforms.keys()}'


    def setup(self, stage: str) -> None:
        """对数据设置转换"""
        self.hparams.max_length = self.get_max_length()
        if self.hparams.transform == 'onerel':
            self.hparams.tag2id = self.onerel_tag2id
            self.hparams.id2tag = {i:l for l,i in self.onerel_tag2id.items()}
        self.hparams.label2id = self.label2id
        self.hparams.id2label = {i:l for l, i in self.label2id.items()}
        # 设置数据转换
        self.dataset.set_transform(transform=self.transforms.get(self.hparams.transform))   

    
    @property
    @lru_cache()
    def label2id(self):
        labels = pd.Series(np.concatenate(self.train_df.triples.values)).apply(lambda x: x['predicate']).drop_duplicates().values
        label2id = {label: i for i, label in enumerate(labels)}
        return label2id
    
        
    def gplinker_transform(self, example) -> Dict:
        batch_text = example['text']
        batch_triples = example['triples']
        batch_inputs = {'input_ids': [], 'attention_mask': [], 'token_type_ids': []}
        batch_so_ids = []
        batch_head_ids = []
        batch_tail_ids = []
        for i, text in enumerate(batch_text):
            inputs = self.tokenizer(text, 
                                    padding='max_length',  
                                    max_length=self.hparams.max_length,
                                    truncation=True,
                                    return_offsets_mapping=True)
            batch_inputs['input_ids'].append(inputs['input_ids'])
            batch_inputs['attention_mask'].append(inputs['attention_mask'])
            batch_inputs['token_type_ids'].append(inputs['token_type_ids'])
            offset_mapping = inputs['offset_mapping']
            triples = batch_triples[i]
            so_ids = torch.zeros(2, self.hparams.max_length, self.hparams.max_length)
            head_ids = torch.zeros(len(self.hparams['label2id']), self.hparams.max_length, self.hparams.max_length)
            tail_ids = torch.zeros(len(self.hparams['label2id']), self.hparams.max_length, self.hparams.max_length)
            for triple in triples:
                try:
                    sub_start = triple['subject']['offset'][0]
                    sub_end = triple['subject']['offset'][1]-1
                    sub_start = char_idx_to_token(sub_start, offset_mapping=offset_mapping)
                    sub_end = char_idx_to_token(sub_end, offset_mapping=offset_mapping)
                    obj_start = triple['object']['offset'][0]
                    obj_end = triple['object']['offset'][1]-1
                    obj_start = char_idx_to_token(obj_start, offset_mapping=offset_mapping)
                    obj_end = char_idx_to_token(obj_end, offset_mapping=offset_mapping)
                    so_ids[0][sub_start][sub_end] = 1
                    so_ids[1][obj_start][obj_end] = 1
                    head_ids[self.hparams['label2id'][triple['predicate']]][sub_start][obj_start] = 1
                    tail_ids[self.hparams['label2id'][triple['predicate']]][sub_end][obj_end] = 1
                except:
                    log.warning(f'sub char offset {(sub_start, sub_end)} or obj char offset {(obj_start, obj_end)} align to token offset failed in \n{text}')
                    pass
            batch_so_ids.append(so_ids)
            batch_head_ids.append(head_ids)
            batch_tail_ids.append(tail_ids)
        batch = dict(zip(batch_inputs.keys(), map(torch.tensor, batch_inputs.values())))
        batch['so_ids'] = torch.stack(batch_so_ids, dim=0)
        batch['head_ids'] = torch.stack(batch_head_ids, dim=0)
        batch['tail_ids'] = torch.stack(batch_tail_ids, dim=0)
        return batch
    
    
    @property
    def onerel_tag2id(self):
        return {'O':0, 'HB-TB':1, 'HB-TE':2, 'HE-TE':3}
    
    
    def onerel_transform(self, example):
        texts = example['text']
        batch_triples = example['triples']
        batch_inputs = {'input_ids': [], 'attention_mask': [], 'token_type_ids': []}
        batch_tag_ids = []
        batch_loss_mask = []
        max_length = self.hparams.max_length
        for i, text in enumerate(texts):
            inputs = self.tokenizer(text, 
                                    padding='max_length',  
                                    max_length=max_length,
                                    truncation=True,
                                    return_offsets_mapping=True)
            batch_inputs['input_ids'].append(inputs['input_ids'])
            batch_inputs['attention_mask'].append(inputs['attention_mask'])
            batch_inputs['token_type_ids'].append(inputs['token_type_ids'])
            offset_mapping = inputs['offset_mapping']
            triples = batch_triples[i]
            tag_ids = torch.zeros(len(self.label2id), max_length, max_length, dtype=torch.long)
            loss_mask = torch.ones(1, max_length, max_length, dtype=torch.long)
            att_mask = torch.tensor(inputs['attention_mask'])
            loss_mask = loss_mask * att_mask.unsqueeze(0) * att_mask.unsqueeze(0).T
            for triple in triples:
                try:
                    rel = triple['predicate']
                    rel_id = self.label2id[rel]
                    sub_start = triple['subject']['offset'][0]
                    sub_end = triple['subject']['offset'][1]-1
                    sub_start = char_idx_to_token(sub_start, offset_mapping=offset_mapping)
                    sub_end = char_idx_to_token(sub_end, offset_mapping=offset_mapping)
                    obj_start = triple['object']['offset'][0]
                    obj_end = triple['object']['offset'][1]-1
                    obj_start = char_idx_to_token(obj_start, offset_mapping=offset_mapping)
                    obj_end = char_idx_to_token(obj_end, offset_mapping=offset_mapping)
                    if sub_start != sub_end and obj_start != obj_end:
                        tag_ids[rel_id][sub_start][obj_start] = self.onerel_tag2id['HB-TB']
                        tag_ids[rel_id][sub_start][obj_end] = self.onerel_tag2id['HB-TE']
                        tag_ids[rel_id][sub_end][obj_end] = self.onerel_tag2id['HE-TE']
                    if sub_start == sub_end and obj_start != obj_end:
                        tag_ids[rel_id][sub_start][obj_start] = self.onerel_tag2id['HB-TB']
                        tag_ids[rel_id][sub_end][obj_end] = self.onerel_tag2id['HE-TE']
                    if sub_start != sub_end and obj_start == obj_end:
                        tag_ids[rel_id][sub_start][obj_start] = self.onerel_tag2id['HB-TB']
                        tag_ids[rel_id][sub_end][obj_end] = self.onerel_tag2id['HE-TE']
                    if sub_start == sub_end and obj_start == obj_end:
                        tag_ids[rel_id][sub_start][obj_start] = self.onerel_tag2id['HB-TB']
                except Exception as e:
                    log.exception(e)
                    log.warning(f'sub char offset {(sub_start, sub_end)} or obj char offset {(obj_start, obj_end)} align to token offset failed in \n{text}')
                    pass
            batch_tag_ids.append(tag_ids)
            batch_loss_mask.append(loss_mask)
        batch = dict(zip(batch_inputs.keys(), map(torch.tensor, batch_inputs.values())))
        batch['tag_ids'] = torch.stack(batch_tag_ids, dim=0)
        batch['loss_mask'] = torch.stack(batch_loss_mask, dim=0)
        return batch
    
    
    def casrel_transform(self, example):
        texts = example['text']
        batch_triples = example['triples']
        batch_inputs = {'input_ids': [], 
                        'attention_mask': [], 
                        'token_type_ids': []}
        batch_subs = []
        batch_objs = []
        batch_sub = []
                        
        max_length = self.hparams.max_length
        for i, text in enumerate(texts):
            inputs = self.tokenizer(text, 
                                    padding='max_length',  
                                    max_length=max_length,
                                    truncation=True,
                                    return_offsets_mapping=True)
            offset_mapping = inputs['offset_mapping']
            triples = batch_triples[i]
            s2ro_map = {}
            for triple in triples:
                sub_start = triple['subject']['offset'][0]
                sub_end = triple['subject']['offset'][1]-1
                sub_start = char_idx_to_token(sub_start, offset_mapping=offset_mapping)
                sub_end = char_idx_to_token(sub_end, offset_mapping=offset_mapping)
                obj_start = triple['object']['offset'][0]
                obj_end = triple['object']['offset'][1]-1
                obj_start = char_idx_to_token(obj_start, offset_mapping=offset_mapping)
                obj_end = char_idx_to_token(obj_end, offset_mapping=offset_mapping)
                rel_id = self.hparams.label2id[triple['predicate']]
                if (sub_start, sub_end) not in s2ro_map:
                    s2ro_map[(sub_start, sub_end)] = [] 
                s2ro_map[(sub_start, sub_end)].append((obj_start, obj_end, rel_id))
                
            if s2ro_map:
                subs = torch.zeros(max_length, 2, dtype=torch.float)
                for s in s2ro_map:
                    subs[s[0], 0] = 1
                    subs[s[1], 1] = 1
                # for sub_head_idx, sub_tail_idx in list(s2ro_map.keys()):
                sub_head_idx, sub_tail_idx = random.choice(list(s2ro_map.keys()))
                sub = torch.tensor([sub_head_idx, sub_tail_idx], dtype=torch.long)
                objs = torch.zeros((max_length, len(self.hparams.label2id), 2), dtype=torch.float)
                for ro in s2ro_map.get((sub_head_idx, sub_tail_idx), []):
                    objs[ro[0], ro[2], 0] = 1
                    objs[ro[1], ro[2], 1] = 1
                batch_subs.append(subs)
                batch_objs.append(objs)
                batch_sub.append(sub)
                batch_inputs['input_ids'].append(inputs['input_ids'])
                batch_inputs['attention_mask'].append(inputs['attention_mask'])
                batch_inputs['token_type_ids'].append(inputs['token_type_ids'])
        batch = dict(zip(batch_inputs.keys(), map(torch.tensor, batch_inputs.values())))
        batch['subs'] = torch.stack(batch_subs, dim=0)
        batch['sub'] = torch.stack(batch_sub, dim=0)
        batch['objs'] = torch.stack(batch_objs, dim=0)
        return batch
        

    @classmethod
    def get_one_example(cls):
        return {'text':'小明是小花的朋友',
                'triples': [{'object': {'offset': [0, 2], 'text': '小明'},
                             'predicate': '朋友',
                             'subject': {'offset': [3, 5], 'text': '小花'}}]}


from __future__ import annotations
from spacy.tokens import Token, Span, Doc, DocBin
from typing import Dict, List, Tuple
import srsly
from spacy.language import Language
from tqdm import tqdm 
import random
import nlhappy


def get_events(doc: Doc) -> List[Event]:
    events=[]
    for event in doc._.event_data:
        label = event['label']
        roles = event['roles']  
        roles = {k:[ doc.char_span(s[0], s[1]) for s in v ] for k, v in roles.items() if len(v)>0}
        if len(roles)>0:
            events.append(Event(label=label, roles=roles))
    return events


def get_relations(doc: Doc) -> List[Relation]:
    """get the relations in the docs"""
    rels = []
    for rel in doc._.rel_data:
        sub = doc.char_span(rel['sub'][0], rel['sub'][1])
        objs = [doc.char_span(obj[0],obj[1]) for obj in rel['objs']]
        if sub is not None and len(objs) > 0 and None not in objs:
            label = rel['label']
            rels.append(Relation(label=label, sub=sub, objs=objs))
    return rels 


Doc.set_extension('rels', getter=get_relations)
Doc.set_extension('events', getter=get_events)
Span.set_extension('norm_name', default='')
Doc.set_extension('label', default='')

# 数据
Doc.set_extension('rel_data', default=[])
Doc.set_extension('event_data',default=[])


class Event():
    def __init__(self,
                 label: str,
                 roles: Dict[str, List[Span]]) -> None:
        """事件类 
        roles:
            label (str): 事件类型
            roles (Dict[str, List[Span]]): 事件中的角色包括触发词
            
        """
        assert isinstance(label, str) and label != '', 'label must be a string, and not empty, but got {}'.format(label)
        assert isinstance(roles, dict) and len(roles) > 0, 'roles must be a dict, and not empty, but got {}'.format(roles)
        self.label = label
        self.roles = roles
        
    @property
    def sents(self):
        left_idx = min([span.start_char for val in self.roles.values() for span in val])
        right_idx = max([span.end_char for val in self.roles.values() for span in val])
        # return self.doc[left_idx:right_idx].sents  ##################### 这个有个bug 暂时不能这样返回
        sents = list(self.doc[left_idx:right_idx].sents)
        last = sents.pop()
        sents.append(last.sent)
        return sents

    @property
    def doc(self) -> Doc:
        for spans in self.roles.values():
            for span in spans:
                return span.doc
        

    def __repr__(self) -> str:
        return f'Event({self.label})'
        
    def __eq__(self, event) -> bool:
        if self.label != event.label:
            return False
        if self.roles.keys() != event.roles.keys():
            return False
        for arg in self.roles.keys():
            if self.roles[arg] != event.roles[arg]:
                return False
        return True
    
    def __hash__(self) -> int:
        return hash(self.__class__)
    
    
class EventData(dict):
    def __init__(self,
                 label: str='',
                 roles: dict = {}):
        super().__init__(label=label, roles=roles)
        self.label = label
        self.roles = roles
        
    def __setitem__(self, key, value):
        assert key in ['label', 'roles'], 'key must be sub, label or objs'
        if key == 'label':
            assert type(value) == str, 'label must be str type'
        if key == 'roles':
            assert type(key) == dict, 'roles must be dict type'
            for role in value:
                assert type(value[role]) == list
                for offset in value[role]:
                    assert offset[0]<offset[1], 'offset start must less than end'
        super().__setitem__(key, value)
        
    
class Relation():
    """used to store the relation in the doc
    """
    def __init__(self,
                 label: str,
                 sub: Span,
                 objs: List[Span]) -> None:
        """
        Args:
            label (str): the relation label
            sub (Span): the subject of the relation
            obj (Span): the object of the relation
        """
        super().__init__()
        assert type(label) == str, 'label must be a string, but got {}'.format(type(label))
        assert type(sub) == Span, 'sub must be a Span, but got {}'.format(type(sub))
        assert type(objs) == list, 'objs must be a list of Span, but got {}'.format(type(objs))
        self.label = label
        self.sub = sub
        self.objs = objs   
          
    @property
    def sents(self) -> List[Span]:
        left_idx = min([self.sub.start_char, min([obj.start_char for obj in self.objs])])
        right_idx = max([self.sub.end_char, max([obj.end_char for obj in self.objs])])
        # return self.doc[left_idx:right_idx].sents
        sents = list(self.doc[left_idx:right_idx].sents)
        last = sents.pop()
        sents.append(last.sent)
        return sents
    
    @property
    def doc(self) -> Doc:
        return self.sub.doc
    
    def __repr__(self) -> str:
        return f'Relation({self.sub},{self.label},{self.objs})'
    
    def __eq__(self, rel) -> bool:
        if rel.label != self.label:
            return False
        if self.sub.start_char != rel.sub.start_char or self.sub.end_char != rel.sub.end_char:
            return False
        if len(self.objs) != len(rel.objs):
            return False 
        for i in range(len(self.objs)):
            if self.objs[i].start_char != rel.objs[i].start_char or self.objs[i].end_char != rel.objs[i].end_char:
                return False
        return True
    
    def __hash__(self) -> int:
        return hash(self.__class__)


class RelationData(dict):
    """存放关系offset数据
    
    args:
        sub(tuple): offset of subject
        label(str): label of relation
        objs(List): offsets of objs
    """
        
    def __init__(self, sub=(), label='', objs=[]):
        super().__init__(sub=sub,label=label,objs=objs)
        self.sub= sub
        self.label = label
        self.objs = objs

    def __setitem__(self, key, value):
        assert key in ['sub', 'label', 'objs'], 'key must be sub, label or objs'
        if key == 'sub':
            assert type(value)== tuple, 'sub must be type tuple'
            assert value[0] < value[1], 'offset start must less than end'
        elif key == 'label':
            assert type(value) == str, 'label must be str type'
        elif key == "objs":
            assert type(value) == list
            for obj in value:
                assert obj[0] < obj[1], 'offset start must less than end'     
        super().__setitem__(key, value)
        
        
def get_docs_from_doccano_jsonl(file_path: str, 
                                 nlp: Language = None, 
                                 set_ent: bool = True,
                                 set_span: bool = False,
                                 set_relation: bool = True) -> List[Doc]:
    """get docs and add entities from doccano jsonl file

    Args:
        file_path (str): the path of the doccano jsonl file
        nlp (Language): the spacy language processing pipeline
        set_ent (bool, optional): whether to set the entities in the doc. Defaults to True.
        set_span (bool, optional): whether to set the spans with a 'all' key in the doc. uses for overlap spans. Defaults to False.
        set_relation (bool, optional): whether to set the relations in the doc. Defaults to True.

    Returns:
        List[Doc]: the list of doc with tags
    """
    if not nlp:
        nlp = nlhappy.nlp()
    docs = []
    for d in tqdm(srsly.read_jsonl(file_path)):
        doc = nlp(d['text'])
        if len(d['entities']) > 0:
            ents = {}
            for ent in d['entities']:
                ents[ent['id']] = doc.char_span(ent['start_offset'], ent['end_offset'], label=ent['label'])
            
            if set_ent:
                try:
                    doc.set_ents(list(ents.values()))
                except:
                    pass
            if set_span:
                doc.spans['all'] = list(ents.values())
            if set_relation:
                if len(d['relations']) > 0:
                    rels = {}
                    for rel in d['relations']:
                        if rel['from_id'] not in rels:
                            rels[rel['from_id']] = {}
                        if rel['type'] not in rels[rel['from_id']]:
                            rels[rel['from_id']][rel['type']] = []
                        rels[rel['from_id']][rel['type']].append(rel['to_id'])
                    for sub_id in rels:
                        for rel_type in rels[sub_id]:
                            sub = ents[sub_id]
                            objs = [ents[obj_id] for obj_id in rels[sub_id][rel_type]]
                            label = rel_type
                            doc._.rel_data.append(RelationData(sub=(sub.start_char, sub.end_char),
                                                          label=label,
                                                          objs=[(obj.start_char, obj.end_char) for obj in objs]))
            
            docs.append(doc)
        
    return docs


def extend_inverse_relations(docs: List[Doc], 
                             inverse_relations: Dict[str, List[str]]) -> List[Doc]:
    """extend the relations in the docs with inverse relations

    Args:
        docs (List[Doc]): the list of docs
        inverse_relations (Dict[str, List[str]]): the inverse relations, key is the relation label, value is the list of inverse relation labels

    Returns:
        List[Doc]: the list of docs with extended relations
    """
    for doc in tqdm(docs):
        for rel in doc._.rel_data:
            if rel.label in inverse_relations:
                for obj in rel.objs:
                    doc._.rel_data.append(RelationData(label=random.sample(inverse_relations[rel.label], k=1)[0], 
                                                  sub=(obj[0], obj[1]), 
                                                  objs=[(rel.sub[0], rel.sub[1])]))
    return docs



def get_docs_from_docbin(db_path: str,
                         nlp: Language = None):
    """get all docs for docbin 

    Args:
        db_path (str): the path to docbin
        nlp (Language, optional): nlp that uses to get vocab. Defaults to None.
    """
    if not nlp:
        nlp = nlhappy.nlp()
    db = DocBin().from_disk(db_path)
    docs = list(db.get_docs(nlp.vocab))
    return docs


import pickle
from ..models import GPLinkerForPromptRelationExtraction, GPLinkerForRelationExtraction
from ..utils.make_doc import RelationData
from spacy.lang.zh import Chinese
from spacy.tokens import Doc
import os
from typing import Union, Optional, List, Tuple
import logging

log = logging.getLogger(__name__)

class GPLinkerExtractor(object):
    '''spacy组件 三元组抽取
    参数:
    - name: 组件名称
    - ckpt: pl模型保存路径
    - model: pl模型名称
    - device: 设备
    '''
    def __init__(self, 
                 nlp, 
                 name:str, 
                 device:str, 
                 threshold:float):
        self.nlp = nlp
        self.pipe_name = name
        self.threshold = threshold
        self.device = device
        
    def __call__(self, doc: Doc) -> Doc:
        triples = self.model.predict(doc.text, self.device, threshold=self.threshold)
        for triple in triples:
            sub_offset = (triple[0], triple[1])
            doc._.rel_data.append(RelationData(sub=sub_offset, label=triple[2], objs=[(triple[3], triple[4])]))
        return doc
    
    def init_model(self, model_or_path):
        if isinstance(model_or_path, GPLinkerForRelationExtraction):
            self.model = model_or_path
            self.model.to(self.device)
            self.model.freeze()
            
        else:
            self.model= GPLinkerForRelationExtraction.load_from_checkpoint(model_or_path)
            self.model.to(self.device)
            self.model.freeze()
    
    def to_disk(self, path:str, exclude):
        # 复制原来模型参数到新的路径
        # path : save_path/information_extractor
        if not os.path.exists(path):
            os.mkdir(path=path)
        model = 're.pkl'
        model_path = os.path.join(path, model)
        with open(model_path, 'wb') as f:
            pickle.dump(self.model, f)
        
    def from_disk(self, path:str, exclude):
        model_path = os.path.join(path, 're.pkl')
        with open(model_path, 'rb') as f:
            self.model = pickle.load(f)
        try:
            self.model.to(self.device)
        except:
            log.info(f' to device {self.device} failed')
        self.model.freeze()
        
class GlobalRelationExtractor():
    '''spacy组件 三元组抽取
    参数:
    - name: 组件名称
    - device: 设备
    '''
    def __init__(self, 
                 nlp, 
                 name:str, 
                 device:str, 
                 threshold:float,
                 prompts:List[str]):
        self.nlp = nlp
        self.pipe_name = name
        self.threshold = threshold
        self.device = device
        self.prompts = prompts
        
    def __call__(self, doc: Doc) -> Doc:
        texts = [doc.text for _ in range(len(self.prompts))]
        triples = self.model.predict(prompts=self.prompts, texts=texts, device=self.device, threshold=self.threshold)
        for triple in triples:
            sub_offset = (triple[0], triple[1])
            doc._.rel_data.append(RelationData(sub=sub_offset, label=triple[2], objs=[(triple[3], triple[4])]))
        return doc
    
    def init_model(self, model_or_path):
        if isinstance(model_or_path, GPLinkerForPromptRelationExtraction):
            self.model = model_or_path
            self.model.to(self.device)
            self.model.freeze()
            
        else:
            self.model= GPLinkerForPromptRelationExtraction.load_from_checkpoint(model_or_path)
            self.model.to(self.device)
            self.model.freeze()
    
    def to_disk(self, path:str, exclude):
        # 复制原来模型参数到新的路径
        # path : save_path/information_extractor
        if not os.path.exists(path):
            os.mkdir(path=path)
        model = 'te.pkl'
        model_path = os.path.join(path, model)
        with open(model_path, 'wb') as f:
            pickle.dump(self.model, f)
        
    def from_disk(self, path:str, exclude):
        model_path = os.path.join(path, 'te.pkl')
        with open(model_path, 'rb') as f:
            self.model = pickle.load(f)
        try:
            self.model.to(self.device)
        except:
            log.info(f' to device {self.device} failed')
        self.model.freeze()
    

default_config = {'threshold':0.0,
                  'prompts':[],
                  'device':'cpu',
                  'model':'global_relation'}

@Chinese.factory('relation_extractor',assigns=['doc._.relations'],default_config=default_config)
def get_relation_extractor(nlp, 
                          name:str, 
                          device:str, 
                          threshold:float,
                          model:str,
                          prompts:List[str]):
    """三元组抽取组件"""
    if model == 'gplinker':
        return GPLinkerExtractor(nlp=nlp, 
                                 name=name, 
                                 device=device, 
                                 threshold=threshold)
    elif model == 'global_relation':
        return GlobalRelationExtractor(nlp=nlp,
                                       name=name,
                                       device=device,
                                       threshold=threshold,
                                       prompts=prompts)


from spacy.tokens import Doc
import os
from typing import Dict, List
from spacy.lang.zh import Chinese
from spacy.language import Language
from ..utils.make_doc import EventData, RelationData
from ..utils.utils import get_logger
from spacy.util import filter_spans
from ..models.prompt_span_extraction import BERTGlobalSpan
import pickle

log = get_logger()


class PSEExtractor:
    def __init__(self, 
                 nlp: Language, 
                 name: str,  
                 device: str,
                 schemas: List[str], 
                 threshold: float =0.5, 
                 set_ents: bool =True,
                 num_sentences: int =0,
                 stride: int =-1):
        self.pipe_name = name
        self.set_ents = set_ents
        self.num_sentences = num_sentences
        self.stride = stride
        self.threshold = threshold
        self.schemas = schemas
        self.device = device

          
        
    def __call__(self, doc: Doc) -> Doc:
        if self.num_sentences > 0:
            raise NotImplementedError
        else:
            #schemas = {'entity':[],'relation':{},'event':{}}
            if 'entity' in self.schemas:
                ent_prompts = self.schemas['entity']
                texts = [doc.text for _ in range(len(ent_prompts))]
                spans = self.model.predict(ent_prompts, texts, device=self.device,threshold=self.threshold)
                if len(spans)>0:
                    doc.spans['all'] = []
                    for span in spans:
                        try:
                            doc.spans['all'].append(doc.char_span(span[0], span[1], label=span[2]))
                        except:
                            log.warning(f'found bad span ({span[0]}, {span[1]}). skip set entity')     
                        if self.set_ents:
                            doc.set_ents(filter_spans(doc.spans['all']))
            if 'relation' in self.schemas:
                sub_prompts = list(self.schemas['relation'].keys())
                texts = texts = [doc.text for _ in range(len(sub_prompts))]
                sub_spans = self.model.predict(sub_prompts, 
                                            texts, 
                                            device=self.device,
                                            threshold=self.threshold)
                if self.set_ents:
                    try:
                        ents = [doc.char_span(s[0], s[1], s[2]) for s in sub_spans]
                        doc.set_ents(filter_spans(ents))
                    except:
                        log.info(f'try set ents on doc({doc.text}) failed')
                for sub_span in sub_spans:
                    sub_label = sub_span[2]
                    sub_text = sub_span[3]
                    rel_types = self.schemas['relation'][sub_label]
                    rel_prompts = [sub_text + '的' + rel_type for rel_type in rel_types]
                    texts = [doc.text for _ in range(len(rel_types))]
                    obj_spans = self.model.predict(rel_prompts, 
                                            texts, 
                                            device=self.device,
                                            threshold=self.threshold)  
                    if len(obj_spans) >0:
                        label_mapping = {}
                        for obj in obj_spans:
                            rel_type = obj[2].split(sub_text)[-1][1:]
                            if rel_type not in label_mapping:
                                label_mapping[rel_type] = []
                            if obj[0]<=obj[1]:
                                label_mapping[rel_type].append((obj[0],obj[1]))
                        sub = (sub_span[0], sub_span[1])
                        for rel_type in label_mapping:
                            obj_offsets = label_mapping[rel_type]
                            doc._.rel_data.append(RelationData(sub, label=rel_type, objs=obj_offsets))
            if 'event' in self.schemas:
                events = list(self.schemas['event'].keys())
                trigger_prompts = [event + '的触发词' for event in events]
                texts = [doc.text for _ in range(len(trigger_prompts))]
                trigger_spans = self.model.predict(trigger_prompts, texts,device=self.device, threshold=self.threshold)
                if len(trigger_spans) > 0:
                    for trigger_span in trigger_spans:
                        event = trigger_span[2].split('的触发词')[0]
                        args = self.schemas['event'][event]
                        args_dict = {}
                        args_dict['触发词'] = [(trigger_span[0], trigger_span[1])]
                        arg_prompts = [event+'的'+arg for arg in args]
                        texts = [doc.text for _ in range(len(arg_prompts))]
                        arg_spans = self.model.predict(prompts=arg_prompts,
                                                       texts=texts,
                                                       device=self.device,
                                                       threshold=self.threshold)
                        for arg_span in arg_spans:
                            arg_type = arg_span[2].split(event+'的')[-1]
                            if arg_type not in args_dict:
                                args_dict[arg_type] = []
                            if arg_span[0] <= arg_span[1]:
                                args_dict[arg_type].append((arg_span[0], arg_span[1]))
                        if len(args_dict)>0:
                            doc._.event_data.append(EventData(label=event, roles=args_dict))         
        return doc
    

    def to_disk(self, path:str, exclude):
        # 复制原来模型参数到新的路径
        # path : save_path/information_extractor
        if not os.path.exists(path):
            os.mkdir(path=path)
        model = 'pse.pkl'
        model_path = os.path.join(path, model)
        with open(model_path, 'wb') as f:
            pickle.dump(self.model, f)
        
    def from_disk(self, path:str, exclude):
        model_path = os.path.join(path, 'pse.pkl')
        with open(model_path, 'rb') as f:
            self.model = pickle.load(f)
        self.model.freeze()
        try:
            self.model.to(self.device)
        except:
            log.info(f' to device {self.device} failed')
        
    def init_model(self, model_or_path):
        if isinstance(model_or_path, BERTGlobalSpan):
            self.model = model_or_path
            self.model.freeze()
            self.model.to(self.device)
            
        else:
            self.model= BERTGlobalSpan.load_from_checkpoint(model_or_path)
            self.model.freeze()
            self.model.to(self.device)
        
default_config = {'threshold': 0.5, 
                  'set_ents': True, 
                  'num_sentences': 0,
                  'stride': -1,
                  'device':'cpu'}
        
        
@Chinese.factory('information_extractor',assigns=['doc.spans','doc.ents','doc._.relations','doc._.events'],default_config=default_config)
def get_information_extractor(nlp: Language,
            name:str, 
            schemas: Dict,  
            num_sentences:bool,
            stride:int, 
            threshold:float, 
            set_ents: bool,
            device:str):
    """the information extractor pipe

    Args:
        nlp (Language): spacy nlp
        name (str): name of the pipe
        schemas (Dict): schemas of the pipe
        num_sentences (bool): num of sentences every time to extract
        threshold (float): threshold of the model
        set_ents (bool): whether to set the doc.ents

    Returns:
        PSEExtractor: extractor based on the prompt-span-extraction model
    """
    
    return PSEExtractor(nlp=nlp, 
                        name=name,
                        schemas=schemas, 
                        num_sentences=num_sentences,
                        stride=stride, 
                        threshold=threshold, 
                        set_ents=set_ents,
                        device=device)
   


import pytorch_lightning as pl
from transformers import AutoModel
import torch.nn as nn
import torch
from ...metrics.span import SpanF1
from ...layers import MultiLabelCategoricalCrossEntropy, EfficientGlobalPointer, MultiDropout
from ...utils.make_model import get_hf_tokenizer
from typing import List


class BERTGlobalSpan(pl.LightningModule):
    
    def __init__(self,
                hidden_size: int,
                lr: float,
                weight_decay: float,
                dropout: float,
                threshold: float = 0.5,
                **kwargs):
        """全局span抽取模型
        Args:
            hidden_size (int): 隐藏层维度
            lr (float): 学习率
            weight_decay (float): 权重衰减
            dropout (float): dropout比例
            threshold (float, optional): 抽取阈值. Defaults to 0.5.
            kwargs: 其他参数包括trf_config, vocab
        """
        super().__init__()
        
        self.save_hyperparameters(logger=False)   


        self.plm = AutoModel.from_config(self.hparams.trf_config)
        self.classifier = EfficientGlobalPointer(
                        input_size=self.plm.config.hidden_size, 
                        hidden_size=hidden_size,
                        output_size=1)

        self.dropout = MultiDropout()
        self.criterion = MultiLabelCategoricalCrossEntropy()

        self.train_metric = SpanF1()
        self.val_metric = SpanF1()
        self.test_metric = SpanF1()
        self.tokenizer = get_hf_tokenizer(config=self.hparams.trf_config, vocab=self.hparams.vocab)

    def forward(self, input_ids, token_type_ids, attention_mask=None):
        x = self.plm(input_ids=input_ids, token_type_ids=token_type_ids, attention_mask=attention_mask).last_hidden_state
        x = self.dropout(x)
        logits = self.classifier(x, mask=attention_mask)
        return logits

    def shared_step(self, batch):
        span_ids = batch['span_ids']
        inputs = batch['inputs']
        logits = self(**inputs)
        pred = logits.ge(self.hparams.threshold).float()
        batch_size, ent_type_size = logits.shape[:2]
        y_true = span_ids.reshape(batch_size*ent_type_size, -1)
        y_pred = logits.reshape(batch_size*ent_type_size, -1)
        loss = self.criterion(y_pred, y_true)
        return loss, pred, span_ids

    def training_step(self, batch, batch_idx):
        loss, pred, true = self.shared_step(batch)
        self.train_metric(pred, true)
        self.log('train/f1', self.train_metric, on_step=True, prog_bar=True)
        self.log('train/loss', loss)
        return {'loss': loss}

    def validation_step(self, batch, batch_idx):
        loss, pred, true = self.shared_step(batch)
        self.val_metric(pred, true)
        self.log('val/f1', self.val_metric, on_epoch=True, prog_bar=True)
        return {'loss': loss}

    def test_step(self, batch, batch_idx):
        loss, pred, true = self.shared_step(batch)
        self.test_metric(pred, true)
        self.log('test/f1', self.test_metric, on_epoch=True, prog_bar=True)
        return {'loss': loss}

    def configure_optimizers(self):
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
        grouped_parameters = [
            {'params': [p for n, p in self.plm.named_parameters() if not any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.plm.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': 0.0},
            {'params': [p for n, p in self.classifier.named_parameters() if not any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr * 5, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.classifier.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr * 5, 'weight_decay': 0.0}
        ]
        optimizer = torch.optim.AdamW(grouped_parameters)
        scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, lambda epoch: 1.0 / (epoch + 1.0))
        return [optimizer], [scheduler]

    def predict(self, prompts: List[str], texts: List[str], device: str='cpu', threshold = None):
        max_length = min(max([len(p+t)+3 for p, t in zip(prompts, texts)]), 512)
        if threshold is None:
            threshold = self.hparams.threshold
        inputs = self.tokenizer(
            prompts,
            texts,
            max_length=max_length,
            truncation=True,
            return_tensors='pt',
            padding='max_length',
            return_offsets_mapping=True)
        mapping = inputs['offset_mapping'].tolist()
        del inputs['offset_mapping']
        inputs.to(device)
        logits = self(**inputs)
        spans_ls = torch.nonzero(logits>threshold).tolist()
        spans = []
        for span in spans_ls :
            start = span[2]
            end = span[3]
            label = prompts[span[0]]
            start_text = mapping[span[0]][start][0]
            end_text = mapping[span[0]][end][1]
            spans.append([start_text, end_text, label, texts[span[0]][start_text:end_text]])
        return spans

    def to_onnx(self, file_path: str):
        prompt = '国家'
        text = '我是中国人'
        torch_inputs = self.tokenizer(prompt, text, return_tensors='pt')
        dynamic_axes = {
                    'input_ids': {0: 'batch', 1: 'seq'},
                    'attention_mask': {0: 'batch', 1: 'seq'},
                    'token_type_ids': {0: 'batch', 1: 'seq'},
                }
        with torch.no_grad():
            torch.onnx.export(
                model=self,
                args=tuple(torch_inputs.values()), 
                f=file_path, 
                input_names=list(torch_inputs.keys()),
                dynamic_axes=dynamic_axes, 
                opset_version=14,
                output_names=['logits'],
                export_params=True)
        print('export to onnx successfully')


from functools import lru_cache
import pytorch_lightning as pl
from typing import Optional,  Tuple, List, Union
from ..utils.make_datamodule import PLMBaseDataModule
import torch
import os


class TextPairClassificationDataModule(PLMBaseDataModule):
    '''句子对数据模块,用来构建pytorch_lightning的数据模块.
    dataset_example:
        {'text_a': '肺结核','text_b': '关节炎','label': 不是一种病}
    '''
    def __init__(self,
                dataset: str,
                plm: str,
                transform: str,
                batch_size: int,
                return_pair: bool=False): 
        """参数:
        - dataset: 数据集名称,feature 必须包含 text_a, text_b, label
        - plm: 预训练模型名称
        - batch_size: 批大小
        - return_pair: 是否以文本对为输入
        """ 
        super().__init__()

        # 这一行代码为了保存传入的参数可以当做self.hparams的属性
        self.save_hyperparameters(logger=False)


    @property
    @lru_cache()
    def label2id(self):
        set_labels = sorted(set([label for label in self.dataset['train']['label']]))
        return {label: i for i, label in enumerate(set_labels)}

        
    def setup(self, stage: str):
        self.hparams['label2id'] = self.label2id
        self.hparams['id2label'] = {i: label for i, label in self.label2id.items()}
        self.dataset.set_transform(transform=self.transform)
        

    def transform(self, examples):
        batch_text_a = examples['text_a']
        batch_text_b = examples['text_b']
        batch_labels = examples['label']
        batch = {'inputs_a': [], 'inputs_b': [], 'label_ids':[]}
        batch_cross = {'inputs': [], 'label_ids':[]}
        if self.hparams.return_pair:
            for i  in range(len(batch_text_a)):
                inputs_a= self.tokenizer(batch_text_a[i], 
                                        padding='max_length', 
                                        max_length=self.hparams.max_length+2, 
                                        truncation=True)
                inputs_a = dict(zip(inputs_a.keys(), map(torch.tensor, inputs_a.values())))
                batch['inputs_a'].append(inputs_a)
                inputs_b = self.tokenizer(batch_text_b[i],
                                        padding='max_length', 
                                        max_length=self.hparams.max_length, 
                                        truncation=True)
                inputs_b = dict(zip(inputs_b.keys(), map(torch.tensor, inputs_b.values())))
                batch['inputs_b'].append(inputs_b)
                batch['label_ids'].append(self.hparams['label2id'][batch_labels[i]])
            return batch
        else:
            for i in range(len(batch_text_a)):
                inputs = self.tokenizer(batch_text_a[i],
                                        batch_text_b[i], 
                                        padding='max_length', 
                                        max_length=self.hparams.max_length*2+2, 
                                        truncation=True)
                inputs = dict(zip(inputs.keys(), map(torch.tensor, inputs.values())))
                batch_cross['inputs'].append(inputs)
                batch_cross['label_ids'].append(self.hparams['label2id'][batch_labels[i]])
            return batch_cross


from spacy.pipeline import Sentencizer
from spacy.lang.zh import Chinese
from spacy.language import Language
from typing import List, Optional, Tuple, Callable
from spacy.pipeline import Pipe
import spacy
import re

default_punct_chars = ['!',  '?', '։', '؟', '۔', '܀', '܁', '܂', '߹',
            '।', '॥', '၊', '။', '።', '፧', '፨', '᙮', '᜵', '᜶', '᠃', '᠉', '᥄',
            '᥅', '᪨', '᪩', '᪪', '᪫', '᭚', '᭛', '᭞', '᭟', '᰻', '᰼', '᱾', '᱿',
            '‼', '‽', '⁇', '⁈', '⁉', '⸮', '⸼', '꓿', '꘎', '꘏', '꛳', '꛷', '꡶',
            '꡷', '꣎', '꣏', '꤯', '꧈', '꧉', '꩝', '꩞', '꩟', '꫰', '꫱', '꯫', '﹒',
            '﹖', '﹗', '！', '．', '？', '𐩖', '𐩗', '𑁇', '𑁈', '𑂾', '𑂿', '𑃀',
            '𑃁', '𑅁', '𑅂', '𑅃', '𑇅', '𑇆', '𑇍', '𑇞', '𑇟', '𑈸', '𑈹', '𑈻', '𑈼',
            '𑊩', '𑑋', '𑑌', '𑗂', '𑗃', '𑗉', '𑗊', '𑗋', '𑗌', '𑗍', '𑗎', '𑗏', '𑗐',
            '𑗑', '𑗒', '𑗓', '𑗔', '𑗕', '𑗖', '𑗗', '𑙁', '𑙂', '𑜼', '𑜽', '𑜾', '𑩂',
            '𑩃', '𑪛', '𑪜', '𑱁', '𑱂', '𖩮', '𖩯', '𖫵', '𖬷', '𖬸', '𖭄', '𛲟', '𝪈',
            '｡', '。', '？', '！', '......', '……', ';', '；', '.']

@Chinese.factory(name='sentence_segmenter', default_config={'punct_chars': default_punct_chars})
def get_sentence_segmenter(nlp, name, punct_chars):
    return Sentencizer(name, punct_chars=punct_chars)




def cut_sentences_v1(sent):
    """
    the first rank of sentence cut
    """
    sent = re.sub('([。？！\?])([^”’])', r"\1\n\2", sent)  # 单字符断句符
    sent = re.sub('(\.{6})([^”’])', r"\1\n\2", sent)  # 英文省略号
    sent = re.sub('(\…{2})([^”’])', r"\1\n\2", sent)  # 中文省略号
    sent = re.sub('([。！？\?][”’])([^，。！？\?])', r"\1\n\2", sent)
    # 如果双引号前有终止符，那么双引号才是句子的终点，把分句符\n放到双引号后
    return sent.split("\n")

def cut_sentences_v2(sent):
    """
    the second rank of spilt sentence, split '；' | ';'
    """
    sent = re.sub('([；;])([^”’])', r"\1\n\2", sent)
    return sent.split("\n")

def cut_sent(text, max_seq_len):
    # 将句子分句，细粒度分句后再重新合并
    sentences = []

    # 细粒度划分
    sentences_v1 = cut_sentences_v1(text)
    for sent_v1 in sentences_v1:
        if len(sent_v1) > max_seq_len - 2:
            sentences_v2 = cut_sentences_v2(sent_v1)
            sentences.extend(sentences_v2)
        else:
            sentences.append(sent_v1)
    assert ''.join(sentences) == text

    # 合并
    merged_sentences = []
    start_index_ = 0

    while start_index_ < len(sentences):
        tmp_text = sentences[start_index_]

        end_index_ = start_index_ + 1

        while end_index_ < len(sentences) and \
                len(tmp_text) + len(sentences[end_index_]) <= max_seq_len - 2:
            tmp_text += sentences[end_index_]
            end_index_ += 1

        start_index_ = end_index_

        merged_sentences.append(tmp_text)

    return merged_sentences


if __name__ == '__main__':
    nlp = spacy.blank('zh')
    nlp.add_pipe('sentence_segmenter')
    text = "肥厚型心肌病@有阵发性或慢性心房颤动的 HCM 患者使用华法林抗凝的国际标准化比值 (INR) 目标值推荐为 2.0-3.0。"
    doc = nlp(text)
    print(list(doc.sents))




import spacy
from typing import Any
from spacy.tokens import Doc
import unicodedata


class CharTokenizer:
    def __init__(self, vocab) -> None:
        self.vocab = vocab

    def __call__(self, text) -> Any:
        words = [s for s in text]
        spaces = [False] * len(words)
        return Doc(self.vocab, words=words, spaces=spaces)
    def to_disk(self, path):
        pass
    
    def from_disk(self, path):
        pass 
    

class BasicTokenizer(object):
    """基础分词器
    - 中文按照字切分
    - 英文按照词切分
    - 数字单个切分
    - 标点切分
    - 英文转小写,去掉重音符号
    - """

    def __init__(self,
                 vocab,
                 do_lower_case=False,
                 return_offset=False,
                 never_split=[]):
        self.do_lower_case = do_lower_case
        self.never_split = never_split
        self.return_offset = return_offset
        self.vocab = vocab

    def tokenize(self, text):
        """文本切分成token"""
        origin = text
        text = self._clean_text(origin)
        text = self._tokenize_chinese_chars(text)
        orig_tokens = self._whitespace_tokenize(text)
        split_tokens = []
        for token in orig_tokens:
            if self.do_lower_case and token not in self.never_split:
                token = token.lower()
                token = self._run_strip_accents(token)
            split_tokens.extend(self._run_split_on_punc(token))

        output_tokens = self._whitespace_tokenize(" ".join(split_tokens))
        if self.return_offset:
            offsets = self.rematch(text=origin, tokens=output_tokens)
            return output_tokens, offsets
        return output_tokens

    def _run_strip_accents(self, text):
        """Strips accents from a piece of text."""
        text = unicodedata.normalize("NFD", text)
        output = []
        for char in text:
            cat = unicodedata.category(char)
            if cat == "Mn":
                continue
            output.append(char)
        return "".join(output)

    def _run_split_on_punc(self, text):
        """Splits punctuation on a piece of text."""
        if text in self.never_split:
            return [text]
        chars = list(text)
        i = 0
        start_new_word = True
        output = []
        while i < len(chars):
            char = chars[i]
            if self._is_punctuation(char):
                output.append([char])
                start_new_word = True
            else:
                if start_new_word:
                    output.append([])
                start_new_word = False
                output[-1].append(char)
            i += 1

        return ["".join(x) for x in output]

    def _tokenize_chinese_chars(self, text:str):
        """Adds whitespace around any CJK character."""
        output = []
        for char in text:
            cp = ord(char)
            if self._is_chinese_char(cp) or char.isdigit():
                output.append(" ")
                output.append(char)
                output.append(" ")
            else:
                output.append(char)
        return "".join(output)

    def _is_chinese_char(self, cp):
        """Checks whether CP is the codepoint of a CJK character."""
        # This defines a "chinese character" as anything in the CJK Unicode block:
        #   https://en.wikipedia.org/wiki/CJK_Unified_Ideographs_(Unicode_block)
        #
        # Note that the CJK Unicode block is NOT all Japanese and Korean characters,
        # despite its name. The modern Korean Hangul alphabet is a different block,
        # as is Japanese Hiragana and Katakana. Those alphabets are used to write
        # space-separated words, so they are not treated specially and handled
        # like the all of the other languages.
        if ((cp >= 0x4E00 and cp <= 0x9FFF) or  #
                (cp >= 0x3400 and cp <= 0x4DBF) or  #
                (cp >= 0x20000 and cp <= 0x2A6DF) or  #
                (cp >= 0x2A700 and cp <= 0x2B73F) or  #
                (cp >= 0x2B740 and cp <= 0x2B81F) or  #
                (cp >= 0x2B820 and cp <= 0x2CEAF) or
                (cp >= 0xF900 and cp <= 0xFAFF) or  #
                (cp >= 0x2F800 and cp <= 0x2FA1F)):  #
            return True

        return False

    def _clean_text(self, text):
        """Performs invalid character removal and whitespace cleanup on text."""
        output = []
        for char in text:
            cp = ord(char)
            if cp == 0 or cp == 0xfffd or self._is_control(char):
                continue
            if self._is_whitespace(char):
                output.append(" ")
            else:
                output.append(char)
        return "".join(output)

    def _whitespace_tokenize(self, text):
        """去除文本中的空白符"""
        text = text.strip()
        if not text:
            return []
        tokens = text.split()
        return tokens



    def _is_whitespace(self, char):
        """Checks whether `chars` is a whitespace character."""
        # \t, \n, and \r are technically contorl characters but we treat them
        # as whitespace since they are generally considered as such.
        if len(char) == 0:
            return False
        if char == " " or char == "\t" or char == "\n" or char == "\r":
            return True
        cat = unicodedata.category(char)
        if cat == "Zs":
            return True
        return False


    def _is_control(self, char):
        """Checks whether `chars` is a control character."""
        # These are technically control characters but we count them as whitespace
        # characters.
        if len(char) ==0:
            return False
        if char == "\t" or char == "\n" or char == "\r":
            return False
        cat = unicodedata.category(char)
        if cat.startswith("C"):
            return True
        return False


    def _is_punctuation(self, char):
        """Checks whether `chars` is a punctuation character."""
        cp = ord(char)
        # We treat all non-letter/number ASCII as punctuation.
        # Characters such as "^", "$", and "`" are not in the Unicode
        # Punctuation class but we treat them as punctuation anyways, for
        # consistency.
        if ((cp >= 33 and cp <= 47) or (cp >= 58 and cp <= 64) or
                (cp >= 91 and cp <= 96) or (cp >= 123 and cp <= 126)):
            return True
        cat = unicodedata.category(char)
        if cat.startswith("P"):
            return True
        return False

    def _convert_to_unicode(self, text):
        """Converts `text` to Unicode (if it's not already), assuming utf-8 input."""
        if isinstance(text, str):
            return text
        elif isinstance(text, bytes):
            return text.decode("utf-8", "ignore")
        else:
            raise ValueError("Unsupported string type: %s" % (type(text)))

    def _lowercase_and_normalize(self, text):
        """转小写，并进行简单的标准化
        """
        text = text.lower()
        text = unicodedata.normalize('NFD', text)
        text = ''.join([ch for ch in text if unicodedata.category(ch) != 'Mn'])
        return text

    def _is_special(self, ch):
        """判断是不是有特殊含义的符号
        """
        return bool(ch) and (ch[0] == '[') and (ch[-1] == ']')

    def rematch(self, text, tokens):
        """给出原始的text和tokenize后的tokens的映射关系
        """

        if self.do_lower_case:
            text = text.lower()

        normalized_text, char_mapping = '', []
        for i, ch in enumerate(text):
            if self.do_lower_case:
                ch = self._lowercase_and_normalize(ch)
            ch = ''.join([
                c for c in ch
                if not (ord(c) == 0 or ord(c) == 0xfffd or self._is_control(c))
            ])
            normalized_text += ch
            char_mapping.extend([i] * len(ch))

        text, token_mapping, offset = normalized_text, [], 0
        for token in tokens:
            if self._is_special(token):
                token_mapping.append(())
            else:
                if self.do_lower_case:
                    token = token.lower()
                start = text[offset:].index(token) + offset
                end = start + len(token)
                # token_mapping.append(char_mapping[start:end])
                token_mapping.append((start, end))
                offset = end

        return token_mapping
    
    def __call__(self, text: str) -> Any:
        tokens = self.tokenize(text)
        spaces = [False] * len(tokens)
        return Doc(self.vocab, words=tokens, spaces=spaces)
    
    def to_disk(self, path):
        pass
    
    def from_disk(self, path):
        pass
    
    
all_tokenizers = {'char':CharTokenizer, 
                 'basic':BasicTokenizer}


from spacy.tokens import Doc
from datasets import Dataset 
from typing import List, Dict, Optional
from tqdm import tqdm 
import random


def convert_spans_to_dataset(docs: List[Doc], 
                             sentence_level: bool =False,
                             user_ents_set_span: bool = True) -> Dataset:
    """
    Convert a document to a dataset.
    args:
        docs: a list of spacy.tokens.Doc
        sentence_level: whether to convert to sentence level dataset
        user_ents_set_span: whether to set spans by using doc.ents
    """

    if sentence_level:
        print('注意: 转换为句子级别数据集, 仅适用于token, span分类任务')
    if isinstance(docs, Doc):
        docs = [docs]
        

    d = {'text':[], 'labels':[],  'spans':[], 'tokens':[]}
    for doc in tqdm(docs, desc='处理数据....'):
        if not sentence_level:
            d['text'].append(doc.text)
            d['labels'].append([label for label in doc._.labels])

            spans = []
            if ('all' in doc.spans) and len(doc.spans['all']) > 0:
                for span in doc.spans['all']:
                    spans.append({'offset': (span.start_char, span.end_char), 'label': span.label_, 'text': span.text})
                
            tokens = []
            if len(doc.ents) > 0:
                for token in doc:
                    t = {'offset': (token.idx, token.idx+1), 'text': token.text}
                    bio = token.ent_iob_ + '-' + token.ent_type_ if token.ent_iob_ != 'O' else token.ent_iob_
                    t['label'] = bio
                    tokens.append(t)
                if user_ents_set_span:
                    for span in doc.ents:
                        spans.append({'offset': (span.start_char, span.end_char), 'label': span.label_, 'text': span.text})
                    
            d['tokens'].append(tokens)
            d['spans'].append(spans)


        else:
            for sent in doc.sents:
                d['text'].append(sent.text)
                d['labels'].append([])
                sent_start = sent.start_char
                spans = []
                if ('all' in doc.spans) and len(doc.spans['all']) > 0:
                    for span in doc.spans['all']:
                        if span.sent == sent:
                            spans.append({'offset': (span.start_char - sent_start, span.end_char - sent_start), 'label': span.label_, 'text': span.text})

                tokens = []
                if len(doc.ents) > 0:
                    for token in sent:
                        t = {'offset': (token.idx - sent_start, token.idx - sent_start+1), 'text': token.text}
                        bio = token.ent_iob_ + '-' + token.ent_type_ if token.ent_iob_ != 'O' else token.ent_iob_
                        t['label'] = bio
                        tokens.append(t)
                    if user_ents_set_span:
                        for span in doc.ents:
                            if span.sent == sent:
                                spans.append({'offset': (span.start_char - sent_start, span.end_char - sent_start), 'label': span.label_, 'text': span.text})
                d['tokens'].append(tokens)
                d['spans'].append(spans)

    print("保存数据....")
    ds = Dataset.from_dict(d)
    return ds




def convert_ents_to_prompt_span_dataset(docs: list,
                                       synonym_dict: Dict[str, List]={}, 
                                       add_negtive_sample: bool = False,
                                       num_samples: int = 1,
                                       sentence_level: bool = False) -> Dataset:
    """将doc转换为prompt span数据集

    Args:
        docs (list): 待转换的文档列表
        synonym_dict (Dict[str, List], optional): 同义词字典. Defaults to {}.
        add_negtive_sample (bool, optional): 是否添加负样本. Defaults to False.
        num_samples (int, optional): 添加负样本的数量. Defaults to 1.
        sentence_level (bool, optional): 是否转换为句子级别. Defaults to False.

    Returns:
        Dataset: 转换的数据集
    """
    all_labels = set([ent.label_ for doc in docs for ent in doc.ents])
    text_ls = []
    prompt_ls = []
    spans_ls = []
    for doc in tqdm(docs):
        if not sentence_level:
            label_dict = {}
            for ent in doc.ents:
                if ent.label_ not in label_dict:
                    label_dict[ent.label_] = []
                label_dict[ent.label_].append({'text': ent.text, 'offset':[ent.start_char, ent.end_char]})
            if len(label_dict) == 0:
                labels = random.choices(list(all_labels),k=num_samples)
                for l in labels:
                    text_ls.append(doc.text)
                    prompt_ls.append(l)
                    spans_ls.append([])
            else:
                for label in label_dict:
                    text_ls.append(doc.text)
                    if label in synonym_dict:
                        prompt_ls.append(random.choice(synonym_dict[label]+[label]))
                    else: 
                        prompt_ls.append(label)
                    spans_ls.append(label_dict[label])
                if add_negtive_sample:
                    other_labels = list(all_labels - set(label_dict.keys()))
                    other_labels = list(filter(None, other_labels))
                    if len(other_labels)>0:
                        random.shuffle(other_labels)
                        for label in other_labels[:num_samples]:
                            text_ls.append(doc.text)
                            if label in synonym_dict:
                                prompt_ls.append(random.choice(synonym_dict[label]+[label]))
                            else:
                                prompt_ls.append(label)
                            spans_ls.append([])
        else:
            for sent in doc.sents:
                label_dict = {}
                for ent in sent.ents:
                    if ent.label_ not in label_dict:
                        label_dict[ent.label_] = []
                    label_dict[ent.label_].append({'text': ent.text, 'offset':[ent.start_char-sent.start_char, ent.end_char-sent.start_char]})
                if len(label_dict) == 0:
                    labels = random.choices(list(all_labels),k=num_samples)
                    for l in labels:
                        text_ls.append(sent.text)
                        prompt_ls.append(l)
                        spans_ls.append([])
                else:
                    for label in label_dict:
                        text_ls.append(sent.text)
                        if label in synonym_dict:
                            prompt_ls.append(random.choice(synonym_dict[label]+[label]))
                        else: 
                            prompt_ls.append(label)
                        spans_ls.append(label_dict[label])
                    if add_negtive_sample:
                        other_labels = list(all_labels - set(label_dict.keys()))
                        other_labels = list(filter(None, other_labels))
                        if len(other_labels)>0:
                            random.shuffle(other_labels)
                            for label in other_labels[:num_samples]:
                                text_ls.append(sent.text)
                                if label in synonym_dict:
                                    prompt_ls.append(random.choice(synonym_dict[label]+[label]))
                                else:
                                    prompt_ls.append(label)
                                spans_ls.append([])   
    print("转换数据....") 
    ds = Dataset.from_dict({'text': text_ls, 'prompt': prompt_ls, 'spans': spans_ls})
    print("转换完成....")
    return ds



    
def convert_events_to_prompt_span_dataset(docs: List[Doc],
                                        synonym_dict: Dict[str, List[str]] = {},
                                        span_synonym_dict : Dict[str, List[str]] = {},
                                        add_negative_sample: bool =False,
                                        negative_roles: int = 1,
                                        negative_events: Optional[int] = None,
                                        sentence_level: bool = False,
                                        add_span_prompt: bool = False,
                                        return_schema: bool = False) -> Dataset:

    """convert doc._.events to prompt span dataset({'text': '', 'prompt': '', 'spans': []})
    Args:
        docs (List[Doc]): doc list
        synonym_dict (Dict[str, List[str]], optional): synonym dict like {'姓名':[名称]}. Defaults to {}.
        span_synonym_dict (Dict[str, List[str]], optional): span synonym dict like {'姓名':[名称]}. Defaults to {}.
        add_negative_sample (bool, optional): add negative sample. Defaults to False.
        negative_roles (int, optional): the num of negative role samples. Defaults to 1.
        negative_events: (int, optinal): the num of negative event samples. if set None that means all sample. Defaults to None
        sentence_level (bool, optional): if cut into sentence-level text. the text contains all sentences the event belongs to. Defaults to False.
        add_span_prompt (bool, optional): if add span prompt to dataset. Defaults to False.

    Returns:
        Dataset: converted dataset
    """
    all_roles = set([role for doc in docs for event in doc._.events for role in event.roles])
    all_span_labels = set([span.label_ for doc in docs for event in doc._.events for role in event.roles for span in event.roles[role]])
    all_event_labels = set([event.label for doc in docs for event in doc._.events])
    schema = {}
    for doc in docs:
        for event in doc._.events:
            if event.label not in schema:
                schema[event.label] = []
            for role in event.roles:
                if role not in schema[event.label]:
                    schema[event.label].append(role)      
    text_ls = []
    prompt_ls = []
    spans_ls = []
    for doc in tqdm(docs):
        doc_events = set([event.label for event in doc._.events])
        for event in doc._.events:
            if not sentence_level:
                span_label_dict = {} # 用来保存每种类别的span
                for role in event.roles:
                    if role in synonym_dict:
                        role = random.choice(synonym_dict[role]+[role])
                    text_ls.append(doc.text)
                    prompt_ls.append(event.label + '的' + role)
                    spans_ls.append([{'text': span.text, 'offset':[span.start_char, span.end_char]} for span in event.roles[role]])
                    if add_span_prompt:
                        for span in event.roles[role]:
                            if span.label_ not in span_label_dict:
                                span_label_dict[span.label_] = []
                            span_label_dict[span.label_].append({'text': span.text, 'offset':[span.start_char, span.end_char]})
                # 添加span prompt数据
                if add_span_prompt:
                    for span_label in span_label_dict:
                        if span_label in span_synonym_dict:
                            span_label = random.choice(span_synonym_dict[span_label]+[span_label])
                        text_ls.append(doc.text)
                        prompt_ls.append(span_label)
                        spans_ls.append(span_label_dict[span_label])
                    
                if add_negative_sample:
                    other_roles = all_roles - set(event.roles.keys())
                    if len(other_roles)>0:
                        other_roles = list(other_roles)
                        random.shuffle(other_roles)
                        other_roles = list(filter(None, other_roles))
                        for role in other_roles[:negative_roles]:
                            if role in synonym_dict:
                                role = random.choice(synonym_dict[role]+[role])
                            text_ls.append(doc.text)
                            prompt_ls.append(event.label + '的' + role)
                            spans_ls.append([])
                    other_event_labels = all_event_labels - doc_events
                    if len(other_event_labels)>0:
                        other_event_labels = list(other_event_labels)
                        random.shuffle(other_event_labels)
                        for label in other_event_labels[:negative_events]:
                            # for i, role in enumerate(event.roles):
                            #     if i == negative_roles:
                            #         break
                            #     if role in synonym_dict:
                            #         role = random.choice(synonym_dict[role]+[role])
                            #     text_ls.append(doc.text)
                            #     prompt_ls.append(label + '的' + role)
                            #     spans_ls.append([]) 
                            # for role in schema[label][:negative_roles]:
                            #     if role in synonym_dict:
                            #         role = random.choice(synonym_dict[role]+[role])
                            #     text_ls.append(doc.text)
                            #     prompt_ls.append(label + '的' + role)
                            #     spans_ls.append([])
                            text_ls.append(doc.text)
                            prompt_ls.append(label + '的' + '触发词')
                            spans_ls.append([])
                            
                                      
                    if add_span_prompt:
                        other_span_labels = all_span_labels - set(span_label_dict.keys())
                        if len(other_span_labels)>0:
                            other_span_labels = list(other_span_labels)
                            random.shuffle(other_span_labels)
                            other_span_labels = list(filter(None, other_span_labels))
                            for span_label in other_span_labels[:negative_roles]:
                                if span_label in span_synonym_dict:
                                    span_label = random.choice(span_synonym_dict[span_label]+[span_label])
                                text_ls.append(doc.text)
                                prompt_ls.append(span_label)
                                spans_ls.append([])
            else:
                text = ''.join([sent.text for sent in event.sents])
                span_label_dict = {} # 用来保存每种类别的span
                for role in event.roles:
                    if role in synonym_dict:
                        role = random.choice(synonym_dict[role]+[role])
                    text_ls.append(text)
                    prompt_ls.append(event.label + '的' + role)
                    spans_ls.append([{'text': span.text, 'offset':[span.start_char-event.sents[0].start_char, span.end_char-event.sents[0].start_char]} for span in event.roles[role]])
                    if add_span_prompt:
                        for span in event.roles[role]:
                            if span.label_ not in span_label_dict:
                                span_label_dict[span.label_] = []
                            span_label_dict[span.label_].append({'text': span.text, 'offset':[span.start_char-event.sents[0].start_char, span.end_char-event.sents[0].start_char]})
                # 添加span prompt数据
                if add_span_prompt:
                    for span_label in span_label_dict:
                        if span_label in span_synonym_dict:
                            span_label = random.choice(span_synonym_dict[span_label]+[span_label])
                        text_ls.append(text)
                        prompt_ls.append(span_label)
                        spans_ls.append(span_label_dict[span_label])
                # 添加负样本 
                if add_negative_sample:
                    other_roles = all_roles - set(event.roles.keys())
                    if len(other_roles)>0:
                        other_roles = list(other_roles)
                        random.shuffle(other_roles)
                        other_roles = list(filter(None, other_roles))
                        for i,role in enumerate(other_roles):
                            if i == negative_roles:
                                break
                            if role in synonym_dict:
                                role = random.choice(synonym_dict[role]+[role])
                            text_ls.append(text)
                            prompt_ls.append(event.label + '的' + role)
                            spans_ls.append([])
                    # other_event_labels = all_event_labels - set(event.label)
                    other_event_labels = all_event_labels - doc_events
                    if len(other_event_labels)>0:
                        other_event_labels = list(other_event_labels)
                        # random.shuffle(other_event_labels)
                        for label in other_event_labels[:negative_events]:
                            # 添加其他事件的当前事件角色负样本
                            # for role in event.roles:
                            #     if role in synonym_dict:
                            #         role = random.choice(synonym_dict[role]+[role])
                            #     text_ls.append(text)
                            #     prompt_ls.append(label + '的' + role)
                            #     spans_ls.append([]) 
                            # # 添加其他事件角色负样本
                            # for role in schema[label][:negative_roles]:
                            #     if role in synonym_dict:
                            #         role = random.choice(synonym_dict[role]+[role])
                            #     text_ls.append(text)
                            #     prompt_ls.append(label + '的' + role)
                            #     spans_ls.append([])
                            text_ls.append(text)
                            prompt_ls.append(label + '的' + '触发词')
                            spans_ls.append([])
                                
                    if add_span_prompt:
                        other_span_labels = all_span_labels - set(span_label_dict.keys())
                        if len(other_span_labels)>0:
                            other_span_labels = list(other_span_labels)
                            random.shuffle(other_span_labels)
                            other_span_labels = list(filter(None, other_span_labels))
                            for span_label in other_span_labels[:negative_roles]:
                                if span_label in span_synonym_dict:
                                    span_label = random.choice(span_synonym_dict[span_label]+[span_label])
                                text_ls.append(text)
                                prompt_ls.append(span_label)
                                spans_ls.append([])
                        
    print("转换为数据集....")
    ds = Dataset.from_dict({'text': text_ls, 'prompt': prompt_ls, 'spans': spans_ls})
    print("转换完成")
    if not return_schema:
        return ds
    else:
        return ds, schema


def convert_relations_to_prompt_span_dataset(docs: List[Doc],
                                             synonym_dict: Dict[str, List[str]]={},
                                             span_synonym_dict: Dict[str, List[str]]={},
                                             sentence_level: bool = False,
                                             add_negative_sample: bool = False,
                                             num_samples: int = 1,
                                             add_span_prompt: bool = False,
                                             add_reverse_negative: bool = True) -> Dataset:
    text_ls = []
    prompt_ls = []
    spans_ls = []
    if add_span_prompt:
        # get all span labels
        all_span_labels = set()
        for doc in docs:
            for rel in doc._.relations:
                if len(rel.sub.ents)==1:
                    all_span_labels.add(rel.sub.ents[0].label_)
                for obj in rel.objs:
                    if len(obj.ents) == 1:
                        all_span_labels.add(obj.ents[0].label_)
    all_relation_labels = set([rel.label for doc in docs for rel in doc._.relations])
    for doc in tqdm(docs):
        if not sentence_level:
            # 添加relation prompt数据
            for rel in doc._.relations:
                text_ls.append(doc.text)
                if rel.label in synonym_dict:
                    rel_label = random.choice(synonym_dict[rel.label]+[rel.label])
                else:
                    rel_label = rel.label
                prompt_ls.append(rel.sub.text + '的' + rel_label)
                spans = []
                for obj in rel.objs:
                    spans.append({'text': obj.text, 'offset':[obj.start_char, obj.end_char]})
                spans_ls.append(spans)
            if add_negative_sample:
                other_rel_labels = all_relation_labels - set([rel.label for rel in doc._.relations])
                if len(other_rel_labels)>0:
                    other_rel_labels = list(other_rel_labels)
                    random.shuffle(other_rel_labels)
                    other_rel_labels = list(filter(None, other_rel_labels))
                    for rel in doc._.relations:
                        for rel_label in other_rel_labels[:num_samples]:
                            if rel_label in synonym_dict:
                                rel_label = random.choice(synonym_dict[rel_label]+[rel_label])
                            text_ls.append(doc.text)
                            prompt_ls.append(rel.sub.text + '的' + rel_label)
                            spans_ls.append([])
                if add_reverse_negative:
                    for obj in rel.objs:
                        text_ls.append(doc.text)
                        if rel.label in synonym_dict:
                            rel_label = random.choice(synonym_dict[rel.label]+[rel.label])
                        else:
                            rel_label = rel.label
                        prompt_ls.append(obj.text + '的' + rel_label)
                        spans_ls.append([])

            # 添加span prompt数据
            if add_span_prompt:
                span_label_dict = {}
                for rel in doc._.relations:
                    if len(rel.sub.ents)==1:
                        sub_label = rel.sub.ents[0].label_
                        if sub_label not in span_label_dict:
                            span_label_dict[sub_label] = []
                        span_label_dict[sub_label].append({'text': rel.sub.text, 'offset':[rel.sub.start_char, rel.sub.end_char]})
                        for obj in rel.objs:
                            if len(obj.ents)==1:
                                obj_label = obj.ents[0].label_
                                if obj_label not in span_label_dict:
                                    span_label_dict[obj_label] = []
                                span_label_dict[obj_label].append({'text': obj.text, 'offset':[obj.start_char, obj.end_char]})
                for span_label in span_label_dict:
                    text_ls.append(doc.text)
                    spans_ls.append(span_label_dict[span_label])
                    if span_label in span_synonym_dict:
                        span_label = random.choice(span_synonym_dict[span_label]+[span_label])
                    prompt_ls.append(span_label)
                    
                if add_negative_sample:
                    other_span_labels = list(all_span_labels - set(span_label_dict.keys()))
                    if len(other_span_labels)>0:
                        random.shuffle(other_span_labels)
                        for span_label in other_span_labels[:num_samples]:
                            text_ls.append(doc.text)
                            if span_label in span_synonym_dict:
                                span_label = random.choice(span_synonym_dict[span_label]+[span_label])
                            
                            prompt_ls.append(span_label)
                            spans_ls.append([])
        else:
            rel_labels = set([rel.label for rel in doc._.relations])
            for rel in doc._.relations:
                text = ''.join([sent.text for sent in rel.sents])
                text_ls.append(text)
                rel_label = rel.label
                if rel_label in synonym_dict:
                    rel_label = random.choice(synonym_dict[rel_label]+[rel_label])
                prompt_ls.append(rel.sub.text + '的' + rel_label)
                spans = []
                for obj in rel.objs:
                    spans.append({'text': obj.text, 'offset':[obj.start_char-rel.sents[0].start_char, obj.end_char-rel.sents[0].start_char]})
                spans_ls.append(spans)
                if add_negative_sample:
                    other_rel_labels = all_relation_labels - rel_labels
                    if len(other_rel_labels)>0:
                        other_rel_labels = list(other_rel_labels)
                        random.shuffle(other_rel_labels)
                        other_rel_labels = list(filter(None, other_rel_labels))
                        for rel_label in other_rel_labels[:num_samples]:
                            if rel_label in synonym_dict:
                                rel_label = random.choice(synonym_dict[rel_label]+[rel_label])
                            text_ls.append(text)
                            prompt_ls.append(rel.sub.text + '的' + rel_label)
                            spans_ls.append([])
                if add_reverse_negative:
                    for obj in rel.objs:
                        text_ls.append(text)
                        if rel.label in synonym_dict:
                            rel_label = random.choice(synonym_dict[rel.label]+[rel.label])
                        else:
                            rel_label = rel.label
                        prompt_ls.append(obj.text + '的' + rel_label)
                        spans_ls.append([])
                        
            if add_span_prompt:
                span_label_dict = {}
                for rel in doc._.relations:
                    if rel.sub.label_ not in span_label_dict:
                        span_label_dict[rel.sub.label_] = []
                    span_label_dict[rel.sub.label_].append({'text': rel.sub.text, 'offset':[rel.sub.start_char, rel.sub.end_char]})
                    for obj in rel.objs:
                        if obj.label_ not in span_label_dict:
                            span_label_dict[obj.label_] = []
                        span_label_dict[obj.label_].append({'text': obj.text, 'offset':[obj.start_char, obj.end_char]})
                for span_label in span_label_dict:
                    left_idx = min([span['offset'][0] for span in span_label_dict[span_label]])
                    right_idx = max([span['offset'][1] for span in span_label_dict[span_label]])
                    sents = list(doc[left_idx:right_idx].sents)
                    last = sents.pop()
                    sents.append(last.sent)
                    text = ''.join([sent.text for sent in sents])
                    text_ls.append(text)
                    spans = []
                    for span in span_label_dict[span_label]:
                        # for span in spans:
                        span_ = {'text': span['text'], 'offset':[span['offset'][0]-sents[0].start_char, span['offset'][1]-sents[0].start_char]}
                        spans.append(span_)
                    spans_ls.append(spans)
                    if span_label in span_synonym_dict:
                        synonym_label = random.choice(span_synonym_dict[span_label]+[span_label])
                        prompt_ls.append(synonym_label)
                    else:
                        prompt_ls.append(span_label)
                    if add_negative_sample:
                        other_span_labels = all_span_labels - set(span_label_dict.keys())
                        if len(other_span_labels)>0:
                            other_span_labels = list(other_span_labels)
                            random.shuffle(other_span_labels)
                            other_span_labels = list(filter(None, other_span_labels))
                            for span_label in other_span_labels[:num_samples]:
                                if span_label in span_synonym_dict:
                                    span_label = random.choice(span_synonym_dict[span_label]+[span_label])
                                text_ls.append(text)
                                prompt_ls.append(span_label)
                                spans_ls.append([])           
    print('转换数据中...')
    assert len(text_ls) == len(prompt_ls) == len(spans_ls)
    ds = Dataset.from_dict({'text': text_ls, 'prompt': prompt_ls, 'spans': spans_ls})
    print('转换数据完成')
    return ds


def convert_relation_to_relation_extraction_dataset(docs: List[Doc], 
                                                    prompt_type: bool=False,
                                                    num_negative: int = 2): 
    """将doc._.relations 转换为relation extraction任务数据

    Args:
        docs (List[Doc]): 待转换docs
        prompt_type (bool, optional): 是否为prompt形式. Defaults to False.

    Returns:
        Dataset: all convert dataset
    """
    if not prompt_type:
        text_ls = []
        triple_ls = []
        for doc in tqdm(docs):
            text_ls.append(doc.text)
            triples = []
            for rel in doc._.relations:
                sub = rel.sub
                pred = rel.label
                for obj in rel.objs:
                    triples.append({'subject':{'offset':(sub.start_char, sub.end_char), 'text':sub.text},
                                    'predicate': pred,
                                    'object':{'offset':(obj.start_char, obj.end_char), 'text':obj.text}})
            triple_ls.append(triples)
        ds = Dataset.from_dict({'text':text_ls, 'triples':triple_ls})
        return ds
    else:
        text_ls = []
        triple_ls = []
        prompt_ls = []
        all_labels = set([rel.label for doc in docs for rel in doc._.relations])
        for doc in tqdm(docs):
            triple_dict = {}
            for rel in doc._.relations:
                sub = rel.sub
                p = rel.label
                if p not in triple_dict:
                    triple_dict[p] = []
                for obj in rel.objs:
                    triple_dict[p].append({'subject':{'offset':(sub.start_char, sub.end_char), 'text':sub.text},
                                    'object':{'offset':(obj.start_char, obj.end_char), 'text':obj.text}})
            
            for p in triple_dict:
                text_ls.append(doc.text)
                triple_ls.append(triple_dict[p])
                prompt_ls.append(p)
            other_labels = list(all_labels - triple_dict.keys())
            for l in other_labels[:num_negative]:
                text_ls.append(doc.text)
                triple_ls.append([])
                prompt_ls.append(l)
                
        ds = Dataset.from_dict({'text':text_ls, 'triples':triple_ls, 'prompts': prompt_ls})
        return ds    


from typing import List

def end_of_chunk(prev_tag, tag, prev_type, type_):
    """Checks if a chunk ended between the previous and current word.
    Args:
        prev_tag: previous chunk tag.
        tag: current chunk tag.
        prev_type: previous type.
        type_: current type.
    Returns:
        chunk_end: boolean.
    """
    chunk_end = False

    if prev_tag == 'S':
        chunk_end = True
    # pred_label中可能出现这种情形
    if prev_tag == 'B' and tag == 'B':
        chunk_end = True
    if prev_tag == 'B' and tag == 'S':
        chunk_end = True
    if prev_tag == 'B' and tag == 'O':
        chunk_end = True
    if prev_tag == 'I' and tag == 'B':
        chunk_end = True
    if prev_tag == 'I' and tag == 'S':
        chunk_end = True
    if prev_tag == 'I' and tag == 'O':
        chunk_end = True

    if prev_tag != 'O' and prev_tag != '.' and prev_type != type_:
        chunk_end = True

    return chunk_end


def start_of_chunk(prev_tag, tag, prev_type, type_):
    """Checks if a chunk started between the previous and current word.
    Args:
        prev_tag: previous chunk tag.
        tag: current chunk tag.
        prev_type: previous type.
        type_: current type.
    Returns:
        chunk_start: boolean.
    """
    chunk_start = False

    if tag == 'B':
        chunk_start = True
    if tag == 'S':
        chunk_start = True

    if prev_tag == 'S' and tag == 'I':
        chunk_start = True
    if prev_tag == 'O' and tag == 'I':
        chunk_start = True

    if tag != 'O' and tag != '.' and prev_type != type_:
        chunk_start = True

    return chunk_start

def convert_tags_to_spans(seq: List[str]):
    """
    将bio或bios标注的序列转换为span形式(label, start, end).
    Args:
        seq (list): sequence of labels.
    Returns:
        list: list of (chunk_type, chunk_start, chunk_end).
    Example:
        seq = ['B-PER', 'I-PER', 'O', 'B-LOC']
        convert_tags_to_spans(seq)
        [('PER', 0, 1), ('LOC', 3, 3)]
    """
    # for nested list
    if any(isinstance(s, list) for s in seq):
        seq = [item for sublist in seq for item in sublist + ['O']]
    prev_tag = 'O'
    prev_type = ''
    begin_offset = 0
    chunks = []
    for i, chunk in enumerate(seq + ['O']):
        tag = chunk[0]
        type_ = chunk.split('-')[-1]

        if end_of_chunk(prev_tag, tag, prev_type, type_):
            chunks.append((prev_type, begin_offset, i - 1))
        if start_of_chunk(prev_tag, tag, prev_type, type_):
            begin_offset = i
        prev_tag = tag
        prev_type = type_

    return chunks


def align_char_span(char_span_offset: List, token_offset_mapping, special_offset=(0,0)) -> List:
    '''对齐字符级别的span标签与bert的子词切分的标签
    参数:
    - char_span_offset: 字符级别的文本片段下标,例如(0,1)
    - token_offset_mapping: 词符与字符下标映射例如[(0,1), (1,2), (2,3) ...]
    输出
    token_span_offset: 词符级别的文本片段下标,例如(0,2)
    '''
    token_span_offset = []
    for i, offset in enumerate(token_offset_mapping):
        if offset != special_offset:
            if offset[0] == char_span_offset[0]:
                start = i
                if offset[1] == char_span_offset[1]:
                    end = i+1
                    break
                else: next
            elif offset[1] == char_span_offset[1]:
                end = i+1
                break
    try:
        if start and end:
            token_span_offset = [start, end]
    except Exception:
        pass
    return token_span_offset


def align_token_span(token_span_offset, token_offset_mapping):
    '''将词符级别的下标对齐为字符级别的下标
    参数
    - token_span_offset: 例如(0, 1) 下标指的是字符的下标
    - token_offset_mapping: 每个词符与字符对应的下标[(0,1),(1,2)]
    返回
    char_span_offset: (0,2)
    '''
    char_span_offset = ()
    if token_span_offset[1] - token_span_offset[0] == 1:
        char_span_offset = token_offset_mapping[token_span_offset[0]]
        return char_span_offset

    else:
        start = token_offset_mapping[token_span_offset[0]][0]
        end = token_offset_mapping[token_span_offset[1]-1][1]
        char_span_offset = (start, end)
        return char_span_offset



import os
from .utils import get_logger
from typing import List, Optional, Dict, Union
from torch.utils.data import DataLoader
from datasets import load_from_disk, load_dataset
import pytorch_lightning as pl
from transformers import AutoConfig, AutoTokenizer, AutoModel
from functools import lru_cache


log = get_logger()

def char_idx_to_token(char_idx, offset_mapping):
    """
    将char级别的idx 转换为token级别的idx
    例如:
    text = '我是中国人'
    tokens = ['我', '是', '中国', '人']
    '国'的字符下标为3, token级别的下标为2
    """
    for index, span in enumerate(offset_mapping):
        if span[0] <= char_idx < span[1]:
            return index
    return -1
        
        
def align_char_span(char_span_offset: tuple, 
                    token_offset_mapping, 
                    special_offset=(0,0),
                    add_special: bool=False) -> tuple:
    '''对齐字符级别的span标签与bert的子词切分的标签
    参数:
    - char_span_offsets: 字符级别的文本片段下标,例如(0,1)
    - token_offset_mapping: 词符与字符下标映射例如[(0,1), (1,2), (2,3) ...]
    输出
    token_span_offset: 词符级别的文本片段下标,例如(0,2)
    '''
    token_span_offset = ()
    bias = 0
    for i, offset in enumerate(token_offset_mapping):
        if offset == special_offset:
            bias += 1
            continue
        if offset != special_offset:
            if offset[0] == char_span_offset[0]:
                start = i
                if offset[1] == char_span_offset[1]:
                    end = i+1
                    break
                else: 
                    continue
            if offset[1] == char_span_offset[1]:
                end = i+1
                break
    try:
        if add_special:
            token_span_offset = (start, end)
        else:
            token_span_offset = (start-bias, end-bias)     
    except:
        log.warning('align offset failed')
    return token_span_offset

def align_char_span_text_b(char_span_offset: tuple, 
                            token_offset_mapping, 
                            special_offset=(0,0),
                            add_pre: bool=True) -> tuple:
    """对齐文本b的token下标, 适用于bert text pair形式

    Args:
        char_span_offset (tuple): 文本中的char级别的offset
        token_offset_mapping (_type_): 映射字典
        special_offset (tuple, optional): 特殊token offset. Defaults to (0,0).
        add_pre (bool, optional): 是否加上之前的token. Defaults to True.

    Returns:
        tuple: token级别的offset
    """
    token_span_offset = ()
    bias = 0
    for i, offset in enumerate(token_offset_mapping):
        if offset == special_offset and i ==0:
            continue
        if offset == special_offset:
            bias = i+1
        if offset != special_offset and bias>0:
            if offset[0] == char_span_offset[0]:
                start = i
                if offset[1] == char_span_offset[1]:
                    end = i+1
                    break
                else: 
                    continue
            if offset[1] == char_span_offset[1]:
                end = i+1
                break
    try:
        if add_pre:
            token_span_offset = (start, end)
        else:
            token_span_offset = (start-bias, end-bias)     
    except:
        log.warning(f'align {char_span_offset} failed')
    return token_span_offset


def prepare_data_from_huffingface(dataset: str,
                             plm: str,
                             dataset_dir: str ='./datasets/',
                             plm_dir: str = './plms/') -> None:
        '''
        下载数据集.这个方法只会在一个GPU上执行一次.
        '''
        dataset_path = os.path.join(dataset_dir, dataset)
        plm_path = os.path.join(plm_dir, plm)
        # 检测数据
        if os.path.exists(dataset_path):
            pass
        else:
            log.info('cannot found dataset in {}.'.format(dataset_path))
            try:
                log.info(f'download dataset {dataset} from huffingface')
                dataset = load_dataset(dataset)
                dataset.save_to_disk(dataset_path)
                log.info(f'download dataset succeed')
            except:
                log.error('download dataset failed')

        if os.path.exists(plm_path):
            pass 
        else : 
            log.info('cannot found plm in {}'.format(plm_path))
            try:
                log.info(f'download plm {plm} from huffingface')
                model = AutoModel.from_pretrained(plm)
                tokenizer = AutoTokenizer.from_pretrained(plm)
                model.save_pretrained(plm_path)
                tokenizer.save_pretrained(plm_path)
                log.info(f'download plm succeed')
            except:
                log.error('download plm failed')



class PLMBaseDataModule(pl.LightningModule):
    """数据模块的基类,子类需要完成setup方法,子类初始化的时候至少包含dataset,plm,batch_size参数,
    内置功能:
    - 自动保存超参数
    - 不同策略自动获取最大文本长度,超过512则取512
    - 下载数据集和预训练模型
    - 自动读取tokenizer
    - 自动读取数据集
    - 自动设置dataloader,数据集需要切分为train,validation,test
    """
    def __init__(self,
                 auto_length: Union[str, int] = 'max',
                 plm_dir: str = 'plms',
                 plm_max_input_length: int = 512,
                 dataset_dir: str = 'datasets',
                 num_workers: int = 0,
                 pin_memory: bool = False,
                 shuffle_train: bool = True,
                 shuffle_val: bool = False,
                 shuffle_test: bool = False):
        super().__init__()
        self.save_hyperparameters()
        self.transforms = {}
    
    
    def prepare_data(self) -> None:
        prepare_data_from_huffingface(dataset=self.hparams.dataset,
                                      plm=self.hparams.plm,
                                      dataset_dir=self.hparams.dataset_dir,
                                      plm_dir=self.hparams.plm_dir)
    
    
    @property
    @lru_cache()
    def dataset(self):
        dataset_path = os.path.join(self.hparams.dataset_dir, self.hparams.dataset)
        dsd = load_from_disk(dataset_path)
        return dsd
            
            
    @lru_cache()
    def get_trf_config(self):
        plm_path = os.path.join(self.hparams.plm_dir, self.hparams.plm)
        config = AutoConfig.from_pretrained(plm_path)
        config = config.to_dict()
        return config
    
    
    @property
    @lru_cache()
    def tokenizer(self):
        plm_path = os.path.join(self.hparams.plm_dir, self.hparams.plm)
        return AutoTokenizer.from_pretrained(plm_path)
    
    
    def get_vocab(self):
        return dict(sorted(self.tokenizer.vocab.items(), key=lambda x: x[1]))


    def get_available_transforms(self):
        return self.transforms.keys()
    
        
    @lru_cache()
    def get_max_length(self, set_to_hparams: bool = True):
        """根据auto_length参数自动获取最大token的长度,并将其添加进hparams
        Args:
            set_to_hparam (bool): 调用该方法时自动设置为self.hparams.max_length,默认为True

        Returns:
            int: 最大token长度
        """
        length = self.train_df.text.map(lambda x: len(self.tokenizer.tokenize(x)))
        if self.hparams.auto_length == 'max':
            max_length = length.max()
        if self.hparams.auto_length == 'mean':
            max_length = int(length.mean())
        if type(self.hparams.auto_length) == int:
            assert self.hparams.auto_length >0, 'max_length length  must > 0'
            max_length = self.hparams.auto_length
        max_length_ = min(self.hparams.plm_max_input_length, max_length+2)
        log.info(f'current max token length: {max_length_}')
        if set_to_hparams:
            self.hparams.max_length = max_length_
        return max_length_
    
    
    @property
    @lru_cache()
    def train_df(self):
        return self.dataset['train'].to_pandas()
    
    
    @property
    @lru_cache()
    def val_df(self):
        return self.dataset['validation'].to_pandas()
    
    
    @property
    @lru_cache()
    def test_df(self):
        return self.dataset['test'].to_pandas()
    
    
    def train_dataloader(self):
        return DataLoader(dataset= self.dataset['train'], 
                          batch_size=self.hparams.batch_size, 
                          num_workers=self.hparams.num_workers, 
                          pin_memory=self.hparams.pin_memory,
                          shuffle=self.hparams.shuffle_train)
    
    
    def val_dataloader(self):
        return DataLoader(dataset=self.dataset['validation'], 
                          batch_size=self.hparams.batch_size, 
                          num_workers=self.hparams.num_workers, 
                          pin_memory=self.hparams.pin_memory,
                          shuffle=self.hparams.shuffle_val)


    def test_dataloader(self):
        return DataLoader(dataset=self.dataset['test'], 
                          batch_size=self.hparams.batch_size, 
                          num_workers=self.hparams.num_workers, 
                          pin_memory=self.hparams.pin_memory,
                          shuffle=self.hparams.shuffle_test)


import pickle
from typing import Dict, List
from ..algorithms.text_match import BM25
from spacy.tokens import Doc
from collections import defaultdict
from spacy.lang.zh import Chinese
from spacy.language import Language
import os
from ..models import BERTCrossEncoder
from ..utils.utils import get_logger

log = get_logger()

class EntityNormalizer:
    def __init__(self,
                nlp,
                name,
                norm_labels: List,
                topk: int =20,
                positive_label: str='1',
                strategy: str='pre',
                threshold: float = 0.5,
                device:str = 'cpu') -> None:
        super().__init__()
        self.name = name
        self.positive_label = positive_label
        self.topk = topk
        self.strategy = strategy
        self.threshold = threshold
        self.norm_labels = norm_labels
        self.device = device
    
    def init_match(self,
                    model_or_path):
        """初始化匹配模型

        Args:
            model_or_path: 
        """
        if isinstance(model_or_path, BERTCrossEncoder):
            self.match_model = model_or_path
            self.match_model.freeze()
            self.match_model.to(self.device)
            
        else:
            self.match_model= BERTCrossEncoder.load_from_checkpoint(model_or_path)
            self.match_model.freeze()
            self.match_model.to(self.device)
        
        
    def init_recall(self,
                    norm_names: List[str],
                    synonym_dict: Dict[str, set]={},
                    tokenizer= None,
                    is_retrain_docs:bool=True,
                    k1: float= 1.5,
                    b: float=0.75,
                    epsilon: float=0.25):
        """初始化bm25模型
        Args:
            norm_names (List[str]): 标准实体列表
            synonym_dict (Dict[str, set], optional): 同义词字典key为别名, value为标准词. Defaults to {}.
            tokenizer (_type_, optional): 分词器,如果为None则为字符切分. Defaults to None.
            is_retrain_docs (bool, optional): 是否保存文档. Defaults to True.
            k1 (float, optional): 词频缩放系数. Defaults to 1.5.
            b (float, optional): 文档长度影响系数,0-1取值,0为完全不影响. Defaults to 0.75.
            epsilon (float, optional): idf的下限值. Defaults to 0.25.
        """
        
        self.map_dict = defaultdict(set)
        for ent in norm_names:
            self.map_dict[ent].add(ent)
        for k, values in synonym_dict.items():
            for v in values:
                if k != v:
                    if k in norm_names:
                        self.map_dict[v].add(k)
                    elif v in norm_names:
                        self.map_dict[k].add(v)
        corpus = list(self.map_dict.keys())       
        self.recall_model = BM25(corpus=corpus,
                                k1=k1,
                                b=b,
                                epsilon=epsilon,
                                tokenizer=tokenizer,
                                is_retain_docs=is_retrain_docs)
            
        
    def normalize(self, query):
        recalls = self.recall_model.recall(query, topk=self.topk) #recalls: [['阿-基综合征', 21.098753076625275], ['阿基综合症', 19.811872460858027]]
        recall_names = [recall[0] for recall in recalls]
        if query in recall_names:
            if len(self.map_dict[query]) == 1:
                return list(self.map_dict[query])[0]
        result_ls = []
        recall_ls = []
        if self.strategy == 'pre':
            recall_ls = [recall[0] for recall in recalls]
            for recall in recall_ls:
                pred, score = self.match_model.predict(text_pair=[query, recall], device=self.device)[0]
                if pred == self.positive_label and score > self.threshold:
                    result_ls.append(recall)
                    break
            if len(result_ls) == 0:
                return ''
            return self.map_dict[result_ls[0]]
        elif self.strategy == 'post': 
            for r in  recalls:
                for v in self.map_dict[r[0]]:
                    if v not in recall_ls:
                        recall_ls.append(v)
            for recall in recall_ls:
                pred, score = self.match_model.predict(text_pair=[query, recall], device=self.device)[0]
                if pred == self.positive_label and score > self.threshold:
                    result_ls.append(recall)
                    break
            if len(result_ls) == 0:
                return ''
            return result_ls[0]
            
    def __call__(self, doc: Doc):
        for ent in doc.ents:
            if ent.label_ in self.norm_labels:
                norm_ = self.normalize(ent.text)
                if type(norm_) == set:
                    assert len(norm_) == 1, f'bad case {norm_}'
                    norm_ = list(norm_)[0]
                ent._.norm_name = norm_
        return doc
    
    def to_disk(self, path:str, exclude):
        # 复制原来模型参数到新的路径
        # path : save_path/entity_normalizer
        if not os.path.exists(path):
            os.mkdir(path=path)
        match_path = os.path.join(path, 'match')
        if not os.path.exists(match_path):
            os.mkdir(match_path)
        match_model_path = os.path.join(match_path, 'bert_tm.pkl')
        with open(match_model_path, 'wb') as f:
            pickle.dump(self.match_model, f)
        recall_path = os.path.join(path, 'recall')
        if not os.path.exists(recall_path):
            os.mkdir(recall_path)
        map_dict = os.path.join(recall_path, 'map_dict.pkl')
        with open(map_dict, 'wb') as f:
            pickle.dump(self.map_dict, f)
        bm25_file = os.path.join(recall_path, 'bm25.pkl')
        with open(bm25_file, 'wb') as f:
            pickle.dump(self.recall_model, f)
        
    def from_disk(self, path:str, exclude):
        # path: load_path/entity_normalizer
        match_model_path = os.path.join(path, 'match', 'bert_tm.pkl')
        with open(match_model_path, 'rb') as f:
            self.match_model = pickle.load(f)
        self.match_model.freeze()
        try:
            self.match_model.to(self.device)
        except:
            log.info(f' to device {self.device} failed')
        recall_path = os.path.join(path, 'recall')
        map_dict_path = os.path.join(recall_path, 'map_dict.pkl')
        self.map_dict = pickle.load(open(map_dict_path, 'rb'))
        bm25_path = os.path.join(recall_path, 'bm25.pkl')
        self.recall_model = pickle.load(open(bm25_path,'rb'))
        

default_config = {
    'topk': 20,
    'positive_label': '1',
    'strategy': 'post',
    'threshold': 0.5,
    'device': 'cpu'
    }
        
        
@Chinese.factory('entity_normalizer',assigns=['span._.norm_name'],default_config=default_config)
def get_entity_normalizer(nlp: Language,
                            name: str,
                            norm_labels: List,
                            topk: int =20,
                            positive_label: str='1',
                            strategy: str='pre',
                            threshold: float = 0.5,
                            device: str = 'cpu'):
    
    return EntityNormalizer(nlp=nlp,
                            name=name,
                            norm_labels=norm_labels,
                            topk=topk,
                            positive_label=positive_label,
                            strategy=strategy,
                            threshold=threshold,
                            device=device)    


from spacy.tokens import Doc
from typing import List
from spacy.training import Example 
from spacy.scorer import get_ner_prf, Scorer



def is_ent_badcase(pred: Doc, true: Doc):
    """判断两个doc是否相等"""
    assert pred.text == true.text
    if len(pred.ents) != len(true.ents):
        return True
    else: 
        pred_offsets = set([(ent.start, ent.end, ent.label_) for ent in pred.ents])
        true_offsets = set([(ent.start, ent.end, ent.label_) for ent in true.ents])
        return len(pred_offsets - true_offsets) != 0 
    
def is_span_badcase(pred: Doc, true: Doc, key: str='all'):
    assert pred.text == true.text
    if len(pred.spans[key]) != len(true.spans[key]):
        return True
    else:
        pred_offsets = set([(ent.start, ent.end, ent.label_) for ent in pred.spans[key]])
        true_offsets = set([(ent.start, ent.end, ent.label_) for ent in true.spans[key]])
        return len(pred_offsets - true_offsets) != 0 


def analysis_ent_type(docs: List[Doc], return_ent_per_type: bool=True, ent_lt:int =100):
    """分析得到各个实体类型所对应的数量
    参数:
    - docs: doc列表
    - return_ent_per_type: 是否返回实体类型数量对应字典,默认返回
    - return_ent_lt: 数量, 返回小于这个数量的实体的doc
    """
    
    ent_types = set([ent.label_ for doc in docs for ent in doc.ents])
    type_dict = dict(zip(ent_types, [0]*len(ent_types)))
    for doc in docs:
        for ent in doc.ents:
            type_dict[ent.label_]+=1
    type_dict = {k:v for k,v  in sorted(type_dict.items(), key=lambda item: item[1])}
    return_types = [k for k,v in type_dict.items() if v < ent_lt]
    return_docs = {k:[] for k in return_types}
    for doc in docs:
        for ent in doc.ents:
            if ent.label_ in return_types:
                return_docs[ent.label_].append(doc)
    if return_ent_per_type:
        return type_dict, return_docs
    else: return return_docs


def analysis_ent_badcase(examples: List[Example], return_prf:bool=True):
    """寻找badcase
    - preds: 预测的doc
    - docs: 真实的doc
    - return_prf: 是否返回每个实体类别的prf值
    """
    badcases = []
    for eg in examples:
        pred = eg.x
        target = eg.y
        if is_ent_badcase(pred, target):
            badcases.append(eg)
    if return_prf:
        prf = get_ner_prf(examples=examples)
        return badcases, prf
    else: return badcases
    

def analysis_span_badcase(examples: List[Example], key: str='all', return_prf: bool=True):
    badcases = []
    if return_prf:
        all_preds = 0
        all_corrects = 0
        all_trues = 0
        per_types = {span.label_ : {'preds':0, 'trues':0, 'corrects':0} for e in examples for span in e.y.spans[key]}
    for eg in examples:
        pred = eg.x
        target = eg.y
        if is_span_badcase(pred, target):
            badcases.append(eg)
        if return_prf:
            all_preds += len(eg.x.spans[key])
            all_trues += len(eg.y.spans[key])
            y_spans = set([(span.start_char, span.end_char, span.label_) for span in eg.y.spans[key]])
            x_spans = set([(span.start_char, span.end_char, span.label_) for span in eg.x.spans[key]])
            corrects = x_spans & y_spans
            all_corrects += len(corrects)
            for span in eg.y.spans[key]:
                per_types[span.label_]['trues'] +=1
            for span in eg.x.spans[key]:
                per_types[span.label_]['preds'] +=1
                span_tuple = (span.start_char, span.end_char, span.label_)
                if span_tuple in y_spans:
                    per_types[span.label_]['corrects'] +=1
    if return_prf:
        span_p = all_corrects/(all_preds + 1e-8)
        span_r = all_corrects/(all_trues + 1e-8)
        span_f = 2*span_p*span_r/(span_p + span_r + 1e-8)
        span_per_types = {l:{} for l in per_types}
        for l, scores in per_types.items():
            p = scores['corrects'] / (scores['preds'] + 1e-8)
            r = scores['corrects'] / (scores['trues'] + 1e-8)
            f = 2 * p * r / (p + r + 1e-8)
            span_per_types[l]['p'] = p
            span_per_types[l]['r'] = r
            span_per_types[l]['f'] = f
            
        scores = {'span_micro_p':span_p, 
                'span_micro_r': span_r,
                'span_micro_f': span_f,
                'span_per_types':span_per_types}
        return badcases, scores
    else:
        return badcases
        
            
    

def analysis_relation_badcase(examples: List[Example] , return_prf:bool=True):
    """错误分析,并且返回细粒度的指标

    Args:
        examples (List[Example]): 一个列表,里面为Example
        return_prf (bool, optional): 是否返回prf分数. Defaults to True.

    Returns:
        List[Examples], Dict : badcase, prf分数
    """
    all_preds = 0
    all_corrects = 0
    all_trues = 0
    per_types = {rel.label : {'preds':0, 'trues':0, 'corrects':0} for e in examples for rel in e.y._.relations}
    badcases = []
    for e in examples:
        corrects = [ rel for rel in e.x._.relations if rel in e.y._.relations]
        if len(corrects) != len(e.y._.relations):
            badcases.append(e)
        if return_prf:
            all_trues += len(e.y._.relations)
            all_preds += len(e.x._.relations)
            all_corrects += len(corrects)
            for rel in e.y._.relations:
                per_types[rel.label]['trues'] +=1
            for rel in e.x._.relations:
                per_types[rel.label]['preds'] +=1
                if rel in e.y._.relations:
                    per_types[rel.label]['corrects'] +=1
    if return_prf:
        rel_p = all_corrects/(all_preds + 1e-8)
        rel_r = all_corrects/(all_trues + 1e-8)
        rel_f = 2*rel_p*rel_r/(rel_p + rel_r + 1e-8)
        rel_per_types = {l:{} for l in per_types}
        for l, scores in per_types.items():
            p = scores['corrects'] / (scores['preds'] + 1e-8)
            r = scores['corrects'] / (scores['trues'] + 1e-8)
            f = 2 * p * r / (p + r + 1e-8)
            rel_per_types[l]['p'] = p
            rel_per_types[l]['r'] = r
            rel_per_types[l]['f'] = f
            
        scores = {'rel_micro_p':rel_p, 
                'rel_micro_r': rel_r,
                'rel_micro_f': rel_f,
                'rel_per_types':rel_per_types}
        return badcases, scores
    else:
        return badcases
    
def analysis_text_badcase(examples: List[Example], return_prf: bool=True):
    all_preds = 0
    all_corrects = 0
    all_trues = 0
    per_types = {e.y._.label : {'preds':0, 'trues':0, 'corrects':0} for e in examples}
    badcases = []
    for e in examples:
        if e.x._.label == e.y._.label and e.x._.label != '':
            all_corrects += 1
        else:
            badcases.append(e)
        if return_prf:
            all_trues += 1
            all_preds += 1
            per_types[e.y._.label]['trues'] +=1
            if e.x._.label != '':
                per_types[e.x._.label]['preds'] +=1
                if e.x._.label == e.y._.label:
                    per_types[e.x._.label]['corrects'] +=1
    if return_prf:
        rel_p = all_corrects/(all_preds + 1e-8)
        rel_r = all_corrects/(all_trues + 1e-8)
        rel_f = 2*rel_p*rel_r/(rel_p + rel_r + 1e-8)
        rel_per_types = {l:{} for l in per_types}
        for l, scores in per_types.items():
            p = scores['corrects'] / (scores['preds'] + 1e-8)
            r = scores['corrects'] / (scores['trues'] + 1e-8)
            f = 2 * p * r / (p + r + 1e-8)
            rel_per_types[l]['p'] = p
            rel_per_types[l]['r'] = r
            rel_per_types[l]['f'] = f
            
        scores = {'text_micro_p':rel_p, 
                'text_micro_r': rel_r,
                'text_micro_f': rel_f,
                'text_per_types':rel_per_types}
        return badcases, scores
    else:
        return badcases


from ...utils.make_model import PLMBaseModel
from ...layers.classifier import EfficientGlobalPointer, GlobalPointer
from ...layers.loss import MultiLabelCategoricalCrossEntropy
from ...layers.dropout import MultiDropout
from ...metrics.event import EventF1, Event, Role
import torch
from itertools import groupby
from typing import Any, Union


class DedupList(list):
    """定义去重的list
    """
    def append(self, x):
        if x not in self:
            super(DedupList, self).append(x)


def clique_search(argus, links):
    """搜索每个节点所属的完全子图作为独立事件
    搜索思路：找出不相邻的节点，然后分别构建它们的邻集，递归处理。
    """
    def neighbors(host, argus, links):
        """构建邻集（host节点与其所有邻居的集合）
        """
        results = [host]
        for argu in argus:
            if host[2:] + argu[2:] in links:
                results.append(argu)
        return list(sorted(results))
    
    Argus = DedupList()
    for i1, (_, _, h1, t1) in enumerate(argus):
        for i2, (_, _, h2, t2) in enumerate(argus):
            if i2 > i1:
                if (h1, t1, h2, t2) not in links:
                    Argus.append(neighbors(argus[i1], argus, links))
                    Argus.append(neighbors(argus[i2], argus, links))
    if Argus:
        results = DedupList()
        for A in Argus:
            for a in clique_search(A, links):
                results.append(a)
        return results
    else:
        return [list(sorted(argus))]

class GPLinkerForEventExtraction(PLMBaseModel):
    def __init__(self, 
                 lr: float = 3e-5,
                 hidden_size: int = 64,
                 weight_decay: float = 0.01,
                 threshold: float = 0.0,
                 scheduler: str = 'harmonic_epoch',
                 **kwargs: Any) -> None:
        super().__init__()
        self.plm = self.get_plm_architecture()

        self.role_classifier = GlobalPointer(input_size=self.plm.config.hidden_size,
                                             hidden_size=hidden_size,
                                             output_size=len(self.hparams.label2id),
                                             add_rope=True)
        self.head_classifier = GlobalPointer(input_size=self.plm.config.hidden_size,
                                             hidden_size=hidden_size,
                                             output_size=1,
                                             add_rope=True)
        self.tail_classifier = GlobalPointer(input_size=self.plm.config.hidden_size,
                                             hidden_size=hidden_size,
                                             output_size=1,
                                             add_rope=True)
        self.dropout = MultiDropout()
        self.role_criterion = MultiLabelCategoricalCrossEntropy()
        self.head_criterion = MultiLabelCategoricalCrossEntropy()
        self.tail_criterion = MultiLabelCategoricalCrossEntropy()
        self.val_metric = EventF1()                  

        
    def forward(self, input_ids, token_type_ids, attention_mask):
        x = self.plm(input_ids=input_ids, token_type_ids=token_type_ids, attention_mask=attention_mask).last_hidden_state
        x = self.dropout(x)
        role_logits = self.role_classifier(x, mask=attention_mask)
        head_logits = self.head_classifier(x, mask=attention_mask)
        tail_logits = self.tail_classifier(x, mask=attention_mask)
        return role_logits, head_logits, tail_logits


    def step(self, batch):
        input_ids = batch['input_ids']
        token_type_ids = batch['token_type_ids']
        attention_mask = batch['attention_mask']
        role_true = batch['role_ids']
        head_true = batch['head_ids']
        tail_true = batch['tail_ids']
        role_logits, head_logits, tail_logits = self(input_ids, token_type_ids, attention_mask)
        role_logits_ = role_logits.reshape(role_logits.shape[0] * role_logits.shape[1], -1)
        role_true_ = role_true.reshape(role_true.shape[0] * role_true.shape[1], -1)
        role_loss = self.role_criterion(role_logits_, role_true_)
        
        head_logits_ = head_logits.reshape(head_logits.shape[0] * head_logits.shape[1], -1)
        head_true_ = head_true.reshape(head_true.shape[0] * head_true.shape[1], -1)
        head_loss = self.head_criterion(head_logits_, head_true_)
        
        tail_logits_ = tail_logits.reshape(tail_logits.shape[0] * tail_logits.shape[1], -1)
        tail_true_ = tail_true.reshape(tail_true.shape[0] * tail_true.shape[1], -1)
        tail_loss = self.tail_criterion(tail_logits_, tail_true_)
        
        loss = role_loss + head_loss + tail_loss
        return loss, role_logits, head_logits, tail_logits


    def training_step(self, batch, batch_idx):
        loss, role_logits, head_logits, tail_logits = self.step(batch)
        self.log('train/loss', loss)
        return {'loss': loss}


    def validation_step(self, batch, batch_idx):
        role_true, head_true, tail_true = batch['role_ids'], batch['head_ids'], batch['tail_ids']
        loss, role_logits, head_logits, tail_logits = self.step(batch)
        batch_events = self.extract_events(role_logits, head_logits, tail_logits, threshold=self.hparams.threshold)
        batch_true_events = self.extract_events(role_true, head_true, tail_true, threshold=self.hparams.threshold)
        self.val_metric(batch_events, batch_true_events)
        self.log('val/f1', self.val_metric, on_step=False, on_epoch=True, prog_bar=True)
        return {'val_loss': loss}


    def extract_events(self, role_logits: torch.Tensor, head_logits: torch.Tensor, tail_logits: torch.Tensor, threshold: Union[None, float]=None):
        if threshold == None:
            threshold = self.hparams.threshold
        role_logits = role_logits.chunk(role_logits.shape[0])
        head_logits = head_logits.chunk(head_logits.shape[0])
        tail_logits = tail_logits.chunk(tail_logits.shape[0])
        assert len(head_logits) == len(tail_logits) == len(role_logits)
        batch_events = []
        for i in range(len(role_logits)):
            roles = []
            for l, h, t in zip(*torch.where(role_logits[i].squeeze(0) > threshold)):
                roles.append(tuple(self.hparams.id2label[l.item()]) + (h.item(), t.item()))
            roles = sorted(set(roles))
            links = set()
            for i1, role1 in enumerate(roles):
                for i2, role2 in enumerate(roles):
                    if i2 > i1:
                        # head_logits[i] 1,1,seq_len, seq_len
                        if head_logits[i][0][0][min(role1[2], role2[2])][max(role1[2], role2[2])] > threshold:
                            if tail_logits[i][0][0][min(role1[3], role2[3])][max(role1[3], role2[3])] > threshold:
                                links.add((role1[2], role1[3], role2[2], role2[3]))
                                links.add((role2[2], role2[3], role1[2], role1[3]))
            events = set()
            for e_label, sub_argus in groupby(roles, key=lambda x: x[0]):
                for event in clique_search(list(sub_argus), links):
                    r = set()
                    for argu in event:
                        start, end = argu[2], argu[3]
                        r.add(Role(start=start, end=end, label=argu[1]))
                    event = Event(label=e_label, roles=r)
                    events.add(event)
            batch_events.append(events)
        return batch_events

    
    def configure_optimizers(self)  :
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
        grouped_parameters = [
            {'params': [p for n, p in self.plm.named_parameters() if not any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.plm.named_parameters() if any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr, 'weight_decay': 0.0},
            {'params': [p for n, p in self.role_classifier.named_parameters() if not any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr* 10, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.role_classifier.named_parameters() if any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr* 10, 'weight_decay': 0.0},
            {'params': [p for n, p in self.head_classifier.named_parameters() if not any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr* 10, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.head_classifier.named_parameters() if any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr* 10, 'weight_decay': 0.0},
            {'params': [p for n, p in self.tail_classifier.named_parameters() if not any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr* 10, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.tail_classifier.named_parameters() if any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr* 10, 'weight_decay': 0.0}
        ]
        optimizer = torch.optim.AdamW(grouped_parameters, eps=1e-5)
        scheduler_config = self.get_scheduler_config(optimizer=optimizer, name=self.hparams.scheduler)
        return [optimizer], [scheduler_config]

    def predict(self):
        pass


from ..utils.make_datamodule import PLMBaseDataModule, get_logger, char_idx_to_token
from functools import lru_cache
import torch
import pandas as pd
import numpy as np


log = get_logger()

class EventExtractionDataModule(PLMBaseDataModule):
    """事件抽取数据模块

    数据集格式:
        {'text':'半年前，患者开始出现持续性头疼',
        'events':[{'label':'临床事件', 'roles':[{'label':'主体词', 'offset':(14,16), 'text':'头疼'},{'label':'描述词', 'offset':(10,13), 'text':'持续性'}]}]}
    """
    def __init__(self, 
                dataset: str, 
                plm: str, 
                batch_size: int,
                auto_length: int = 'max',
                transform: str = 'gplinker',
                **kwargs):
        super().__init__()
        self.transforms = {'gplinker': self.gplinker_transform}


    def setup(self, stage: str = 'fit') -> None:
        self.hparams.max_length = self.get_max_length()
        if self.hparams.transform == 'gplinker':
            self.hparams.id2label = {i:l for l,i in self.combined2id.items()}
            self.dataset.set_transform(self.gplinker_transform)


    @property
    @lru_cache()
    def event2id(self):
        labels = pd.Series(np.concatenate(self.train_df.events.values)).apply(lambda x: x['label']).drop_duplicates().values
        return {l:i for i, l in enumerate(labels)}

    @property
    @lru_cache()
    def role2id(self):
        labels = pd.Series(np.concatenate(pd.Series(np.concatenate(self.train_df.events.values)).apply(lambda x: x['roles']).values)).apply(lambda x: x['label']).drop_duplicates().values
        return {l:i for i, l in enumerate(labels)}

    @property
    @lru_cache()
    def combined2id(self):
        def get_labels(e):
            labels = []
            label = e['label']
            for r in e['roles']:
                labels.append((label,r['label']))
            return labels
        df = pd.DataFrame(np.concatenate(pd.Series(np.concatenate(self.train_df.events.values)).apply(get_labels).values).tolist()).drop_duplicates()
        labels = sorted([tuple(v) for v in df.values])
        return {l:i for i,l in enumerate(labels)}



    def gplinker_transform(self, examples):
        batch_text = examples['text']
        batch_events = examples['events']
        batch_inputs = {'input_ids':[], 
                        'token_type_ids':[],
                        'attention_mask':[]}
        batch_role_ids = []
        batch_head_ids = []
        batch_tail_ids = []
        for i, text in enumerate(batch_text):
            inputs = self.tokenizer(text, 
                                    padding='max_length',  
                                    max_length=self.hparams.max_length,
                                    truncation=True,
                                    return_offsets_mapping=True)
            batch_inputs['input_ids'].append(inputs['input_ids'])
            batch_inputs['attention_mask'].append(inputs['attention_mask'])
            batch_inputs['token_type_ids'].append(inputs['token_type_ids'])
            offset_mapping = inputs['offset_mapping']
            events = batch_events[i]
            role_ids = torch.zeros(len(self.combined2id), self.hparams.max_length, self.hparams.max_length)
            head_ids = torch.zeros(1, self.hparams.max_length, self.hparams.max_length)
            tail_ids = torch.zeros(1, self.hparams.max_length, self.hparams.max_length)
            for event in events:
                e_label = event['label']
                for i, role in enumerate(event['roles']):
                    label = (e_label, role['label'])
                    try:
                        start_ = role['offset'][0]
                        end_ = role['offset'][1]-1
                        start = char_idx_to_token(start_, offset_mapping=offset_mapping)
                        end = char_idx_to_token(end_, offset_mapping=offset_mapping)
                    except:
                        log.warning(f'role {role["text"]} offset {(start, end)} align to token offset failed in \n{text}')
                        continue
                    role_ids[self.combined2id[label]][start][end] = 1
                        # 这个role 跟 其他的每个role 头头 尾尾 联系起来
                    for j, role1 in enumerate(event['roles']):
                        if j>i:
                            start1 = role1['offset'][0]
                            end1 = role1['offset'][1]-1
                            try:
                                start1 = char_idx_to_token(start1, offset_mapping=offset_mapping)
                                end1 = char_idx_to_token(end1, offset_mapping=offset_mapping)
                            except:
                                log.warning(f'role {role1["text"]} offset {(start1, end1)} align to token offset failed in \n{text}')
                                continue
                            head_ids[0][min(start, start1)][max(start, start1)] = 1
                            tail_ids[0][min(end, end1)][max(end, end1)] = 1
            batch_role_ids.append(role_ids)
            batch_head_ids.append(head_ids)
            batch_tail_ids.append(tail_ids)
        batch = dict(zip(batch_inputs.keys(), map(torch.tensor, batch_inputs.values())))
        batch['role_ids'] = torch.stack(batch_role_ids, dim=0)
        batch['head_ids'] = torch.stack(batch_head_ids, dim=0)
        batch['tail_ids'] = torch.stack(batch_tail_ids, dim=0)
        return batch


    @staticmethod
    def get_one_example():
        return {'text':'半年前，患者开始出现持续性头疼',
                'events':[{'label':'临床事件', 'roles':[{'label':'主体词', 'offset':(14,16), 'text':'头疼'},
                                                       {'label':'描述词', 'offset':(10,13), 'text':'持续性'}]}]}